#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from onnxruntime.quantization.quant_utils import QuantizationMode
from .operators.base_operator import QuantOperatorBase
from .operators.qdq_base_operator import QDQOperatorBase
from .operators.matmul import MatMulInteger, QLinearMatMul, QDQMatMul
from .operators.attention import AttentionQuant, QDQAttention
from .operators.embed_layernorm import EmbedLayerNormalizationQuant
from .operators.gather import GatherConverter, GatherQuant
from .operators.conv import QLinearConv, ConvInteger, QDQConv
from .operators.activation import QLinearActivation, QDQRemovableActivation, QDQActivation
from .operators.binary_op import QLinearBinaryOp, QDQBinaryOp
from .operators.maxpool import QMaxPool, QDQMaxPool
from .operators.gavgpool import QGlobalAveragePool
from .operators.lstm import LSTMQuant, QDQLSTM
from .operators.split import QSplit, QDQSplit
from .operators.concat import QLinearConcat, QDQConcat
from .operators.pad import QPad, QDQPad
from .operators.pooling import QLinearPool, QDQPool
from .operators.direct_q8 import QDQDirect8BitOp, Direct8BitOp, DirectCast
from .operators.base_operator_cast import CastOperatorBase

CommonOpsRegistry = {"Gather": GatherConverter, \
                     "EmbedLayerNormalization": EmbedLayerNormalizationQuant}

IntegerOpsRegistry = {
    "Conv": ConvInteger,
    "FusedConv": ConvInteger, 
    "MatMul": MatMulInteger,
    "Attention": AttentionQuant,
    "LSTM": LSTMQuant,
}
IntegerOpsRegistry.update(CommonOpsRegistry)

QLinearOpsRegistry = {
    "Conv": QLinearConv,
    "Concat": QLinearConcat,
    "Attention": AttentionQuant,
    "FusedConv": QLinearConv,
    "MatMul": QLinearMatMul,
    "Add": QLinearBinaryOp,
    "Mul": QLinearBinaryOp,
    "Relu": QLinearActivation,
    "Clip": QLinearActivation,
    "LeakyRelu" : QLinearActivation,
    "Sigmoid" : QLinearActivation,
    "MaxPool": QMaxPool,
    "GlobalAveragePool": QGlobalAveragePool,
    "Split": QSplit,
    "Pad": QPad,
    "AveragePool" : QLinearPool,
    "Reshape": Direct8BitOp,
    "Transpose" : Direct8BitOp,
    "Squeeze" : Direct8BitOp,
    "Unsqueeze" : Direct8BitOp,
}
QLinearOpsRegistry.update(CommonOpsRegistry)

QDQRegistry = {
    "FusedConv": QDQConv,
    "Conv": QDQConv,
    "Clip": QDQRemovableActivation,
    "Relu": QDQRemovableActivation,
    "LeakyRelu": QDQActivation,
    "Sigmoid": QDQActivation,
    "MaxPool": QDQMaxPool,
    "MatMul": QDQMatMul,
    "Add": QDQBinaryOp,
    "Mul": QDQBinaryOp,
    "Gather": GatherQuant,
    "Attention": QDQAttention,
    "LSTM": QDQLSTM,
    "Pad": QDQPad,
    "Reshape": QDQDirect8BitOp,
    "Transpose" : QDQDirect8BitOp,
    "Squeeze" : QDQDirect8BitOp,
    "AveragePool": QDQPool,
    "Unsqueeze" : QDQDirect8BitOp,
    "Concat": QDQConcat,
    "Split": QDQSplit
}

CastRegistry = {
    "Shape": DirectCast,
    "Squeeze": DirectCast,
    "Unsqueeze": DirectCast,
    "Reshape": DirectCast,
    "Unsqueeze": DirectCast,
    "Transpose": DirectCast,
    "Loop": DirectCast,
    "Slice": DirectCast,
    "Split": DirectCast,
    "Concat": DirectCast,

}


def CreateOpConverter(onnx_quantizer, node):
    registry = IntegerOpsRegistry if onnx_quantizer.mode == QuantizationMode.IntegerOps \
                                  else QLinearOpsRegistry
    if node.op_type in registry.keys():
        return registry[node.op_type](onnx_quantizer, node)
    return QuantOperatorBase(onnx_quantizer, node)

def CreateQDQQuantizer(onnx_quantizer, node):
    if node.op_type in QDQRegistry.keys():
        return QDQRegistry[node.op_type](onnx_quantizer, node)
    return QDQOperatorBase(onnx_quantizer, node)

def CreateCaster(onnx_quantizer, node):
    if node.op_type in CastRegistry.keys():
        return CastRegistry[node.op_type](onnx_quantizer, node)
    return CastOperatorBase(onnx_quantizer, node)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from .utils import logger
from .data import DATALOADERS, DATASETS
from .experimental import Quantization as ExpQuantization

class Quantization(object):
    """Quantization class automatically searches for optimal quantization recipes for low
       precision model inference, achieving best tuning objectives like inference performance
       within accuracy loss constraints.

       Tuner abstracts out the differences of quantization APIs across various DL frameworks
       and brings a unified API for automatic quantization that works on frameworks including
       tensorflow, pytorch and mxnet.

       Since DL use cases vary in the accuracy metrics (Top-1, MAP, ROC etc.), loss criteria
       (<1% or <0.1% etc.) and tuning objectives (performance, memory footprint etc.).
       Tuner class provides a flexible configuration interface via YAML for users to specify
       these parameters.

    Args:
        conf_fname_or_obj (string or obj): The path to the YAML configuration file or 
            Quantization_Conf class containing accuracy goal, tuning objective and preferred
            calibration & quantization tuning space etc.

    """

    def __init__(self, conf_fname_or_obj):
        self.exp_quantizer = ExpQuantization(conf_fname_or_obj)

    def __call__(self, model, q_dataloader=None, q_func=None, eval_dataloader=None,
                 eval_func=None):
        """The main entry point of automatic quantization tuning.

           This interface works on all the DL frameworks that neural_compressor supports
           and provides three usages:
           a) Fully yaml configuration: User specifies all the info through yaml,
              including dataloaders used in calibration and evaluation phases
              and quantization tuning settings.

              For this usage, only model parameter is mandatory.

           b) Partial yaml configuration: User specifies dataloaders used in calibration
              and evaluation phase by code.
              The tool provides built-in dataloaders and evaluators, user just need provide
              a dataset implemented __iter__ or __getitem__ methods and invoke dataloader()
              with dataset as input parameter to create neural_compressor dataloader before calling this
              function.

              After that, User specifies fp32 "model", calibration dataset "q_dataloader"
              and evaluation dataset "eval_dataloader".
              The calibrated and quantized model is evaluated with "eval_dataloader"
              with evaluation metrics specified in the configuration file. The evaluation tells
              the tuner whether the quantized model meets the accuracy criteria. If not,
              the tuner starts a new calibration and tuning flow.

              For this usage, model, q_dataloader and eval_dataloader parameters are mandatory.

           c) Partial yaml configuration: User specifies dataloaders used in calibration phase
              by code.
              This usage is quite similar with b), just user specifies a custom "eval_func"
              which encapsulates the evaluation dataset by itself.
              The calibrated and quantized model is evaluated with "eval_func".
              The "eval_func" tells the tuner whether the quantized model meets
              the accuracy criteria. If not, the Tuner starts a new calibration and tuning flow.

              For this usage, model, q_dataloader and eval_func parameters are mandatory.

        Args:
            model (object):                        For Tensorflow model, it could be a path
                                                   to frozen pb,loaded graph_def object or
                                                   a path to ckpt/savedmodel folder.
                                                   For PyTorch model, it's torch.nn.model
                                                   instance.
                                                   For MXNet model, it's mxnet.symbol.Symbol
                                                   or gluon.HybirdBlock instance.
            q_dataloader (generator):              Data loader for calibration, mandatory for
                                                   post-training quantization. It is iterable
                                                   and should yield a tuple (input, label) for
                                                   calibration dataset containing label,
                                                   or yield (input, _) for label-free calibration
                                                   dataset. The input could be a object, list,
                                                   tuple or dict, depending on user implementation,
                                                   as well as it can be taken as model input.
            q_func (function, optional):           Training function for Quantization-Aware
                                                   Training. It is optional and only takes effect
                                                   when user choose "quant_aware_training"
                                                   approach in yaml.
                                                   This function takes "model" as input parameter
                                                   and executes entire training process with self
                                                   contained training hyper-parameters. If this
                                                   parameter specified, eval_dataloader parameter
                                                   plus metric defined in yaml, or eval_func
                                                   parameter should also be specified at same time.
            eval_dataloader (generator, optional): Data loader for evaluation. It is iterable
                                                   and should yield a tuple of (input, label).
                                                   The input could be a object, list, tuple or
                                                   dict, depending on user implementation,
                                                   as well as it can be taken as model input.
                                                   The label should be able to take as input of
                                                   supported metrics. If this parameter is
                                                   not None, user needs to specify pre-defined
                                                   evaluation metrics through configuration file
                                                   and should set "eval_func" paramter as None.
                                                   Tuner will combine model, eval_dataloader
                                                   and pre-defined metrics to run evaluation
                                                   process.
            eval_func (function, optional):        The evaluation function provided by user.
                                                   This function takes model as parameter,
                                                   and evaluation dataset and metrics should be
                                                   encapsulated in this function implementation
                                                   and outputs a higher-is-better accuracy scalar
                                                   value.

                                                   The pseudo code should be something like:

                                                   def eval_func(model):
                                                        input, label = dataloader()
                                                        output = model(input)
                                                        accuracy = metric(output, label)
                                                        return accuracy

        Returns:
            quantized model: best qanitized model found, otherwise return None

        """

        logger.warning("This API is going to be deprecated. Please import "
            "neural_compressor.experimental.Quantization, initialize an instance of `Quantization`,"
            "set its dataloader and metric attributes, then invoke its __call__ method.")

        self.exp_quantizer.model = model
        if q_dataloader is not None:
            self.exp_quantizer.calib_dataloader = q_dataloader
        elif q_func is not None:
            self.exp_quantizer.q_func = q_func

        if eval_func is not None:
            self.exp_quantizer.eval_func = eval_func 
        elif eval_dataloader is not None:
            self.exp_quantizer.eval_dataloader = eval_dataloader

        nc_model = self.exp_quantizer.fit()
        if self.exp_quantizer.framework == 'tensorflow':
            return nc_model.graph if nc_model else None
        if self.exp_quantizer.framework == 'pytorch':
            saved_path = os.path.abspath(os.path.join(os.path.expanduser(
                self.exp_quantizer.conf.usr_cfg.tuning.workspace.path), 'checkpoint'))
            nc_model.save(saved_path)
        return nc_model.model

    fit = __call__

    def dataset(self, dataset_type, *args, **kwargs):
        return DATASETS(self.exp_quantizer.framework)[dataset_type](*args, **kwargs)

    def dataloader(self, dataset, batch_size=1, collate_fn=None, last_batch='rollover',
                   sampler=None, batch_sampler=None, num_workers=0, pin_memory=False):
        return DATALOADERS[self.exp_quantizer.framework](
            dataset=dataset,
            batch_size=batch_size, collate_fn=collate_fn, last_batch=last_batch,
            sampler=sampler, batch_sampler=batch_sampler, num_workers=num_workers,
            pin_memory=pin_memory
        )

    def metric(self, name, metric_cls, **kwargs):
        from .experimental.common import Metric as NCMetric
        nc_metric = NCMetric(metric_cls, name, **kwargs)
        self.exp_quantizer.metric = nc_metric

    def postprocess(self, name, postprocess_cls, **kwargs):
        from .experimental.common import Postprocess as NCPostprocess
        nc_postprocess = NCPostprocess(postprocess_cls, name, **kwargs)
        self.exp_quantizer.postprocess = nc_postprocess


def fit(model, conf, calib_dataloader=None, eval_func=None):
    quantizer = Quantization(conf)
    quantizer.model = model
    quantizer(model, calib_dataloader, eval_func)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math


class KL_Divergence(object):
    '''The class of supporting KL divergence calibration algorithm.

       Args:
           TODO:
    '''

    def __init__(self):
        pass

    def expand_quantized_bins(self, quantized_bins, reference_bins):
        """expand quantized bins"""
        expanded_quantized_bins = [0] * len(reference_bins)
        num_merged_bins = int(len(reference_bins) / len(quantized_bins))
        j_start = 0
        j_end = num_merged_bins
        for idx in range(len(quantized_bins)):
            zero_count = reference_bins[j_start:j_end].count(0)
            num_merged_bins = j_end - j_start
            if zero_count == num_merged_bins:
                avg_bin_ele = 0
            else:
                avg_bin_ele = quantized_bins[idx] / (num_merged_bins -
                                                     zero_count + 0.0)
            for idx1 in range(j_start, j_end):
                expanded_quantized_bins[
                    idx1] = 0 if reference_bins[idx1] == 0 else avg_bin_ele
            j_start += num_merged_bins
            j_end += num_merged_bins
            if idx + 1 == len(quantized_bins) - 1:
                j_end = len(reference_bins)
        return expanded_quantized_bins

    def safe_entropy(self, reference_distr_P, P_sum, candidate_distr_Q, Q_sum):
        """ safe entropy """
        assert len(reference_distr_P) == len(candidate_distr_Q)
        tmp_sum1 = 0
        tmp_sum2 = 0
        for idx in range(len(reference_distr_P)):
            p_idx = reference_distr_P[idx]
            q_idx = candidate_distr_Q[idx]
            if p_idx == 0:
                tmp_sum1 += 0
                tmp_sum2 += 0
            else:
                if q_idx == 0:
                    print("Fatal error!, idx = " + str(idx) +
                          " qindex = 0! p_idx = " + str(p_idx))
                tmp_sum1 += p_idx * (math.log(Q_sum * p_idx))
                tmp_sum2 += p_idx * (math.log(P_sum * q_idx))
        return (tmp_sum1 - tmp_sum2) / P_sum

    def get_threshold(self,
                      hist,
                      hist_edges,
                      min_val,
                      max_val,
                      num_bins,
                      quantized_type,
                      num_quantized_bins=255):
        '''The interface of getting threshold per KL divergency algorithm.

           Args:
               historgram (tensor list): The tensor numpy array list including all ops
                                         in each iteration.
               quantized_type (string): string being "int8" or "uint8".
               number_bins (integer): number of bins used in KL.

           Return:
               threshold per quantized ops, it's scalar.
        '''

        if min_val >= 0:
            ending_iter = num_bins - 1
            starting_iter = int(ending_iter * 0.7)
        else:
            th = max(abs(max_val), abs(min_val))
            starting_iter = 0
            ending_iter = num_bins - 1
            if abs(max_val) > abs(min_val):
                while starting_iter < ending_iter:
                    if hist[starting_iter] == 0:
                        starting_iter += 1
                        continue
                    else:
                        break
                starting_iter += int((ending_iter - starting_iter) * 0.6)
            else:
                while ending_iter > 0:
                    if hist[ending_iter] == 0:
                        ending_iter -= 1
                        continue
                    else:
                        break
                starting_iter = int(0.6 * ending_iter)

        bin_width = hist_edges[1] - hist_edges[0]
        min_kl_divergence = 0
        min_kl_index = 0
        kl_inited = False

        for i in range(starting_iter, ending_iter + 1):
            reference_distr_P = hist[0:i].tolist()
            outliers_count = sum(hist[i:2048])
            if reference_distr_P[i - 1] == 0:
                continue
            reference_distr_P[i - 1] += outliers_count
            reference_distr_bins = reference_distr_P[:]
            candidate_distr_Q = hist[0:i].tolist()
            num_merged_bins = int(i / num_quantized_bins)
            candidate_distr_Q_quantized = [0] * num_quantized_bins
            j_start = 0
            j_end = num_merged_bins

            for idx in range(num_quantized_bins):
                candidate_distr_Q_quantized[idx] = sum(
                    candidate_distr_Q[j_start:j_end])
                j_start += num_merged_bins
                j_end += num_merged_bins
                if idx + 1 == num_quantized_bins - 1:
                    j_end = i
            candidate_distr_Q = self.expand_quantized_bins(
                candidate_distr_Q_quantized, reference_distr_bins)
            P_sum = sum(reference_distr_P)
            Q_sum = sum(candidate_distr_Q)
            kl_divergence = self.safe_entropy(reference_distr_P, P_sum,
                                              candidate_distr_Q, Q_sum)
            if not kl_inited:
                min_kl_divergence = kl_divergence
                min_kl_index = i
                kl_inited = True
            elif kl_divergence < min_kl_divergence:
                min_kl_divergence = kl_divergence
                min_kl_index = i
            else:
                pass

        if min_kl_index == 0:
            while starting_iter > 0:
                if hist[starting_iter] == 0:
                    starting_iter -= 1
                    continue
                else:
                    break
            min_kl_index = starting_iter
        return (min_kl_index + 0.5) * bin_width



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neural_compressor.experimental.metric import METRICS
from neural_compressor.experimental.data import DATASETS, TRANSFORMS, FILTERS, DATALOADERS
from neural_compressor.experimental.common import Optimizers, Criterions
from collections import OrderedDict
import copy
import gc

DEFAULT_BATCH_SIZE = 64


def get_func_from_config(func_dict, cfg, compose=True):
    func_list = []
    for func_name, func_value in OrderedDict(cfg).items():
        func_kwargs = {}
        func_args = []
        if isinstance(func_value, dict):
            func_kwargs = func_value
        elif func_value is not None:
            func_args.append(func_value)
        func_list.append(func_dict[func_name](*func_args, **func_kwargs))

    func = func_dict['Compose'](func_list) if compose else \
        (func_list[0] if len(func_list) > 0 else None)
    return func


def get_preprocess(preprocesses, cfg, compose=True):
    return get_func_from_config(preprocesses, cfg, compose)


def get_metrics(metrics, cfg, compose=True):
    return get_func_from_config(metrics, cfg, compose)

def get_postprocess(postprocesses, cfg, compose=True):
    return get_func_from_config(postprocesses, cfg, compose)

def get_algorithm(algorithms, cfg, compose=False):
    # recipes contains quantization part, only use algorithms in that
    algo_conf = algorithms.support_algorithms().intersection(set(cfg.keys()))
    #(TODO) only support open/close according to cfg
    return [algorithms()[algo]() for algo in algo_conf if cfg[algo]]

def create_dataset(framework, data_source, cfg_preprocess, cfg_filter):
    transform_list = []
    # generate framework specific transforms
    preprocess = None
    if cfg_preprocess is not None:
        preprocesses = TRANSFORMS(framework, 'preprocess')
        preprocess = get_preprocess(preprocesses, cfg_preprocess)
    # even we can unify transform, how can we handle the IO, or we do the transform here
    datasets = DATASETS(framework)
    dataset_type = list(data_source.keys())[0]
    # generate framework and dataset specific filters
    filter = None
    if cfg_filter is not None:
        filters = FILTERS(framework)
        filter_type = list(cfg_filter.keys())[0]
        filter_dataset_type = filter_type + dataset_type
        filter = filters[filter_dataset_type](**cfg_filter[filter_type])
    # in this case we should prepare eval_data and calib_data sperately
    dataset = datasets[dataset_type](**data_source[dataset_type],
                                     transform=preprocess, filter=filter)
    return dataset


def create_dataloader(framework, dataloader_cfg):
    batch_size = int(dataloader_cfg['batch_size']) \
        if dataloader_cfg.get('batch_size') is not None else DEFAULT_BATCH_SIZE
    last_batch = dataloader_cfg['last_batch'] \
        if dataloader_cfg.get('last_batch') is not None else 'rollover'
    shuffle = dataloader_cfg['shuffle'] \
        if dataloader_cfg.get('shuffle') is not None else False
    distributed = dataloader_cfg['distributed'] \
        if dataloader_cfg.get('distributed') is not None else False

    dataset = create_dataset(framework,
                             copy.deepcopy(dataloader_cfg['dataset']),
                             copy.deepcopy(dataloader_cfg['transform']),
                             copy.deepcopy(dataloader_cfg['filter']),)

    return DATALOADERS[framework](dataset=dataset,
                                  batch_size=batch_size,
                                  last_batch=last_batch,
                                  shuffle=shuffle,
                                  distributed=distributed)


def create_eval_func(framework, dataloader, adaptor,
                     metric, postprocess_cfg=None,
                     iteration=-1, tensorboard=False,
                     fp32_baseline=False):
    """The interface to create evaluate function from config.

    Args:
        model (object): The model to be evaluated.

    Returns:
        Objective: The objective value evaluated
    """

    # eval_func being None means user will provide dataloader and metric info
    # in config yaml file
    assert dataloader, "dataloader should NOT be empty when eval_func is None"
    postprocess = None
    if postprocess_cfg is not None:
        postprocesses = TRANSFORMS(framework, "postprocess")
        postprocess = get_postprocess(postprocesses, postprocess_cfg['transform'])
    if isinstance(metric, dict):
        fwk_metrics = METRICS(framework)
        metrics = []
        for name, val in metric.items():
            if isinstance(val, int) and \
                len([i for i in gc.get_objects() if id(i) == val]) > 0 and \
                'user_' + type([i for i in gc.get_objects() if id(i) == val][0]).__name__ == name:
                metrics.extend([i for i in gc.get_objects() if id(i) == val])
            elif name not in ['weight', 'higher_is_better']:
                metrics.append(get_metrics(fwk_metrics, {name: val}, compose=False))
    else:
        metrics = metric

    def eval_func(model, measurer=None):
        return adaptor.evaluate(model, dataloader, postprocess,
                                metrics, measurer, iteration,
                                tensorboard, fp32_baseline)
    # TODO: to find a better way
    eval_func.builtin = True

    return eval_func


def create_train_func(framework, dataloader, adaptor, train_cfg, hooks=None, callbacks=None):
    """The interface to create train function from config.

    Args:
        framework (str): The string of framework.
        dataloader (common.DataLoader): The object of common.DataLoader.
        adaptor (obj): The object of adaptor.
        train_cfg (dict): The dict of training related config.
        hooks (dict): The dict of training hooks, supported keys are:
                      on_epoch_begin, on_epoch_end, on_step_begin, on_step_end.
                      Their values are functions to be executed in adaptor layer.

    Returns:
        The constructed train function
    """
    # train_func being None means user will provide training related info
    # in config yaml file
    assert dataloader, "dataloader should NOT be empty when train_func is None"
    assert adaptor, "adaptor should NOT be empty"
    postprocess_cfg = train_cfg.postprocess
    if postprocess_cfg is not None:
        postprocesses = TRANSFORMS(framework, "postprocess")
        postprocess = get_postprocess(postprocesses, postprocess_cfg['transform'])
    else:
        postprocess = None
    if isinstance(train_cfg.optimizer, dict):
        assert train_cfg.optimizer and len(train_cfg.optimizer) == 1, \
            "optimizer should only set once"
        key, value = next(iter(train_cfg.optimizer.items()))
        optimizer = Optimizers(framework)[key](value)
        optimizer = optimizer()
    else:
        if framework == "pytorch":
            optimizer = (lambda mp, p: train_cfg.optimizer, {'p':0})
        elif framework == 'tensorflow':
            optimizer = (lambda p: train_cfg.optimizer, {'p':0})

    if isinstance(train_cfg.criterion, dict):
        assert train_cfg.criterion and len(train_cfg.criterion) == 1, \
            "criterion should only set once"
        key, value = next(iter(train_cfg.criterion.items()))
        criterion = Criterions(framework)[next(iter(train_cfg.criterion))](value)
        criterion = criterion()
    else:
        criterion = (lambda p: train_cfg.criterion, {'p':0})

    default_dict = {k: train_cfg[k] for k in train_cfg.keys() - {'optimizer', 'criterion',
                                                                 'dataloader'}}
    default_dict['callbacks'] = callbacks

    def train_func(model):
        return adaptor.train(model, dataloader, optimizer_tuple=optimizer,
                             criterion_tuple=criterion, hooks=hooks,
                             postprocess=postprocess, kwargs=default_dict)
    # TODO: to find a better way
    train_func.builtin = True

    return train_func



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Metric(object):
    """common Metric just collect the infos to construct a Metric
    """
    def __init__(self, metric_cls, name='user_metric', **kwargs):
        """The metric class should take the outputs of the model as the metric's inputs,
           neural_compressor built-in metric always take (predictions, labels) as inputs, it's
           recommended to design metric_cls to take (predictions, labels) as inputs.
           metric_cls should be sub_class of neural_compressor.metric.BaseMetric.
        """
        self.metric_cls = metric_cls
        self.name = name
        self.kwargs = kwargs




#
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from neural_compressor.adaptor.pytorch import (
    PyTorch_FXAdaptor, 
    get_torch_version, 
    PyTorchVersionMode
    )

version = get_torch_version()

def trace_and_fuse_sub_graph(model, prefix, is_qat):
    from torch.quantization.quantize_fx import _fuse_fx
    import torch.quantization.quantization_mappings as tqqm
    fx_white_list = tqqm.get_default_qconfig_propagation_list()
    for name, module in model.named_children():
        # FX QAT cannot fallback nn.Dropout from train mode to eval
        if type(module) == torch.nn.Dropout:  # pragma: no cover
            continue
        op_name = prefix + '.' + name if prefix != '' else name
        if type(module) in fx_white_list:
            module = torch.quantization.QuantWrapper(module)
        if PyTorch_FXAdaptor._check_dynamic_control(module):
            trace_and_fuse_sub_graph(module, op_name, is_qat=is_qat)
        else:
            try:
                graph_module = torch.fx.symbolic_trace(module)
                if version >= PyTorchVersionMode.PT111.value:  # pragma: no cover
                    fused_model = _fuse_fx(graph_module, is_qat)
                else:
                    fused_model = _fuse_fx(graph_module)  # pylint: disable=E1120
                setattr(model, name, fused_model)
            except:
                trace_and_fuse_sub_graph(module, op_name, is_qat)
    return model

def symbolic_trace(model, is_qat=False):
    try:
        traced_model = torch.fx.symbolic_trace(model)
    except:
        traced_model = trace_and_fuse_sub_graph(model, prefix='', is_qat=is_qat)
    return traced_model


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from neural_compressor.utils.utility import combine_histogram

class LayerHistogramCollector(object):
    """Saves layer histogram in a dict with layer names as keys and lists of NDArrays as
    values. The collected histogram will be used for calculating the optimal thresholds for
    quantization using KL divergence.
    """

    def __init__(self, num_bins=8001, layer_tensor=None, include_layer=None, logger=None):
        self.hist_dict = {}
        self.num_bins = num_bins
        self.layer_tensor = layer_tensor
        self.include_layer = include_layer
        self.logger = logger

    def collect(self):
        """Callback function for collecting layer output NDArrays."""
        # name = py_str(name)
        # if name not in self.include_layer:
        #     return
        # handle = ctypes.cast(arr, NDArrayHandle)
        # arr = NDArray(handle, writable=False).copyto(cpu()).asnumpy()
        for name in self.layer_tensor:
            if name not in self.include_layer:
                continue
            for arr in self.layer_tensor[name]:
                if self.logger:
                    self.logger.debug(
                        "Collect layer {} histogram of shape {}.".format(name, arr.shape))
                min_range = np.min(arr)
                max_range = np.max(arr)
                th = max(abs(min_range), abs(max_range))
                if name in self.hist_dict:
                    self.hist_dict[name] = combine_histogram(self.hist_dict[name], arr)
                else:
                    hist, hist_edges = np.histogram(
                        arr, bins=self.num_bins, range=(-th, th))
                    self.hist_dict[name] = (
                        hist, hist_edges, min_range, max_range, th)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=no-member

import os
import copy
import logging
from collections import OrderedDict
from collections.abc import KeysView
import yaml
import numpy as np
from distutils.version import StrictVersion
from neural_compressor.adaptor.adaptor import adaptor_registry, Adaptor
from neural_compressor.adaptor.query import QueryBackendCapability
from neural_compressor.utils.utility import LazyImport, dump_elapsed_time, \
                                            GLOBAL_STATE, MODE
from neural_compressor.utils.utility import Statistics
from neural_compressor.experimental.data.dataloaders.base_dataloader import BaseDataLoader
from neural_compressor.conf.dotdict import deep_get
import math

onnx = LazyImport("onnx")
ort = LazyImport("onnxruntime")
ort_ext = LazyImport("onnxruntime_extensions")
ONNXRT152_VERSION = StrictVersion("1.5.2")
ONNXRT170_VERSION = StrictVersion("1.7.0")

logger = logging.getLogger()

class ONNXRTAdaptor(Adaptor):
    """The ONNXRT adaptor layer, do onnx-rt quantization, calibration, inspect layer tensors.

    Args:
        framework_specific_info (dict): framework specific configuration for quantization.
    """

    def __init__(self, framework_specific_info):
        super().__init__(framework_specific_info)
        self.__config_dict = {}
        self.quantizable_ops = []
        self.device = framework_specific_info["device"]
        self.static = framework_specific_info["approach"] == "post_training_static_quant"
        self.backend = framework_specific_info["backend"]
        self.work_space = framework_specific_info["workspace_path"]
        self.graph_optimization = framework_specific_info["graph_optimization"]
        self.recipes = deep_get(framework_specific_info, 'recipes', {})
        self.benchmark = (GLOBAL_STATE.STATE == MODE.BENCHMARK)
        os.makedirs(self.work_space, exist_ok=True)
        self.pre_optimized_model = None
        self.quantizable_op_types = []
        for precision in self.query_handler.get_precisions():
            if precision != 'fp32':
                if self.device == 'cpu' and precision == 'fp16':
                    continue
                self.quantizable_op_types += \
                    self.query_handler.get_op_types_by_precision(precision=precision)
 
        self.evaluate_nums = 0

        self.fp32_results = []
        self.fp32_preds_as_label = False
        self.quantize_config = {} # adaptor should know current configs at any time
        self.quantize_params = {} # adaptor should know current params at any time
        self.min_max = None

    @dump_elapsed_time("Pass quantize model")
    def quantize(self, tune_cfg, model, data_loader, q_func=None):
        """The function is used to do calibration and quanitization in post-training
           quantization.

        Args:
            tune_cfg (dict):     quantization config.
            model (object):      model need to do quantization.
            data_loader (object): calibration dataset.
            q_func (optional):   training function for quantization aware training mode,
                                 unimplement yet for onnx.

        Returns:
            (dict): quantized model
        """
        assert q_func is None, "quantization aware training has not been supported on ONNXRUNTIME"
        model = self.pre_optimized_model if self.pre_optimized_model else model
        ort_version = StrictVersion(ort.__version__)
        if ort_version < ONNXRT152_VERSION: # pragma: no cover
            logger.warning("Quantize input needs onnxruntime 1.5.2 or newer.")
            return model
        if model.model.opset_import[0].version < 11: # pragma: no cover
            logger.warning("Quantize input needs model opset 11 or newer.")
        from onnxruntime.quantization.quant_utils import QuantizationMode
        if self.backend in ["qlinearops", "qoperator"]:
            backend = QuantizationMode.QLinearOps
            if self.backend == "qlinearops":
                logger.warning("onnxrt_qlinearops uses the same model representation format as "
                               "onnxrt_qoperator. Recommended to use onnxrt_qoperator to align "
                               "with ONNXRUNTIME QuantFormat")
        elif self.backend == "qdq":
            assert ort_version >= ONNXRT170_VERSION, 'QDQ mode needs onnxruntime1.7.0 or newer'
            backend = "qdq"
        else:
            backend = QuantizationMode.IntegerOps

        self.quantizable_ops = self._query_quantizable_ops(model.model)
        tmp_model = copy.deepcopy(model)

        quantize_config = self._cfg_to_quantize_config(tune_cfg)
        iterations = tune_cfg.get('calib_iteration', 1)
        calib_sampling_size = tune_cfg.get('calib_sampling_size', 1)
        if self.static:
            if isinstance(data_loader, BaseDataLoader):
                batch_size = data_loader.batch_size
                try:
                    for i in range(batch_size):
                        if calib_sampling_size % (batch_size - i) == 0:
                            calib_batch_size = batch_size - i
                            if i != 0:  # pragma: no cover
                                logger.warning("Reset `calibration.dataloader.batch_size` field "
                                               "to {}".format(calib_batch_size) +
                                               " to make sure the sampling_size is "
                                               "divisible exactly by batch size")
                            break
                    tmp_iterations = int(math.ceil(calib_sampling_size / calib_batch_size))
                    data_loader.batch(calib_batch_size)
                    quantize_params = self._get_quantize_params(tmp_model.model, data_loader, \
                                                                quantize_config, tmp_iterations)
                except Exception:  # pragma: no cover
                    logger.warning(
                        "Fail to forward with batch size={}, set to {} now.".
                        format(batch_size, 1))
                    data_loader.batch(1)
                    quantize_params = self._get_quantize_params(tmp_model.model, data_loader, \
                                                                quantize_config, calib_sampling_size)
            else:  # pragma: no cover
                if hasattr(data_loader, 'batch_size') and \
                  calib_sampling_size % data_loader.batch_size != 0:
                    logger.warning(
                        "Please note that calibration sampling size {} " \
                        "isn't divisible exactly by batch size {}. " \
                        "So the real sampling size is {}.".
                        format(calib_sampling_size, data_loader.batch_size,
                               data_loader.batch_size * iterations))
                quantize_params = self._get_quantize_params(tmp_model.model, data_loader, \
                                                            quantize_config, iterations)
        else:
            quantize_params = None
        self.quantize_params = quantize_params
        from neural_compressor.adaptor.ox_utils.quantizer import Quantizer
        quantizer = Quantizer(tmp_model.model,
            quantize_config,
            backend,
            self.static,
            quantize_params,
            self.quantizable_op_types,
            self.query_handler.get_fallback_list())
        quantizer.quantize_model()
        tmp_model.q_config = self._generate_qconfig(model.model, tune_cfg, quantize_params)
        tmp_model.model = quantizer.model.model
        self.quantize_config = quantize_config # update so other methods can know current configs

        self._dump_model_op_stats(tmp_model)
        tmp_model.topological_sort()
        return tmp_model

    def _generate_qconfig(self, model, tune_cfg, quantize_params):
        tune_cfg = copy.deepcopy(tune_cfg)
        for node in model.graph.node:
            if (node.name, node.op_type) not in tune_cfg['op']:
                continue
            scale_info = {}
            if quantize_params:
                for input_name in node.input:
                    if input_name in quantize_params:
                        scale_info[input_name] = quantize_params[input_name]
                for output_name in node.output:
                    if output_name in quantize_params:
                        scale_info[output_name] = quantize_params[output_name]
            tune_cfg['op'][(node.name, node.op_type)]['scale_info'] = scale_info
        fwk_info = {}
        fwk_info['approach'] = "post_training_static_quant" if self.static else \
                                                        "post_training_dynamic_quant"
        fwk_info['backend'] = self.backend
        fwk_info['workspace_path'] = self.work_space
        fwk_info['graph_optimization'] = self.graph_optimization
        fwk_info['device'] = self.device
        tune_cfg['framework_specific_info'] = fwk_info
        return tune_cfg

    @dump_elapsed_time("Pass recover model")
    def recover(self, model, q_config):
        """Execute the recover process on the specified model.

        Args:
            model (object):  model need to do quantization.
            q_config (dict): recover configuration

        Returns:
            (dict): quantized model
        """
        self._pre_optimize(model)
        model = self.pre_optimized_model
        ort_version = StrictVersion(ort.__version__)
        if ort_version < ONNXRT152_VERSION: # pragma: no cover
            logger.warning("Quantize input needs onnxruntime 1.5.2 or newer.")
            return model
        if model.model.opset_import[0].version < 11: # pragma: no cover
            logger.warning("Quantize input needs model opset 11 or newer.")

        from onnxruntime.quantization.quant_utils import QuantizationMode
        if self.backend in ["qlinearops", "qoperator"]:
            backend = QuantizationMode.QLinearOps
        elif self.backend == "qdq":
            assert ort_version >= ONNXRT170_VERSION, 'QDQ mode needs onnxruntime1.7.0 or newer'
            backend = self.backend
        else:
            backend = QuantizationMode.IntegerOps
        from neural_compressor.adaptor.ox_utils.quantizer import Quantizer
        self.quantizable_ops = self._query_quantizable_ops(model.model)
        quantize_params, tune_cfg = self._parse_qconfig(q_config)
        quantize_config = self._cfg_to_quantize_config(tune_cfg)
        quantizer = Quantizer(model.model,
            quantize_config,
            backend,
            self.static,
            quantize_params,
            self.quantizable_op_types,
            self.query_handler.get_fallback_list())
 
        quantizer.quantize_model()
        model.model = quantizer.model.model
        model.topological_sort()
        return model

    def _parse_qconfig(self, q_config):
        quantize_params = {}
        tune_cfg = {}
        for k, v in q_config.items():
            if k == 'op':
                tune_cfg['op'] = {}
                for op_name_type, op_info in v.items():
                    node_dict = {}
                    for info_name, info_content in op_info.items():
                        if info_name != 'scale_info':
                            node_dict[info_name] = info_content
                        else:
                            for tensor_name, param in info_content.items():
                                quantize_params[tensor_name] = param
                    tune_cfg['op'][op_name_type] = node_dict
            else:
                tune_cfg[k] = v
        if len(quantize_params) == 0:
            quantize_params = None
        return quantize_params, tune_cfg

    def _dump_model_op_stats(self, model):
        fp32_op_list = []
        for precision in self.query_handler.get_precisions():
            if precision != 'fp32':
                fp32_op_list += self.query_handler.get_op_types_by_precision(precision=precision)
        res = {}
        for op_type in fp32_op_list:
            res[op_type] = {'INT8':0, 'BF16': 0, 'FP16': 0, 'FP32':0}
        for op_type in ["QuantizeLinear", "DequantizeLinear", "DynamicQuantizeLinear"]:
            res[op_type] = {'INT8':0, 'BF16': 0, 'FP16': 0, 'FP32':0}


        if self.backend in ["qlinearops", "qdq", "qoperator"] :
            int8_op_list = ["QLinearConv", "QLinearMatMul", "QAttention",
                            "QLinearMul", "QLinearRelu", "QLinearClip",
                            "QLinearLeakyRelu", "QLinearSigmoid", "MaxPool","Squeeze",
                            "EmbedLayerNormalization", "QLinearGlobalAveragePool", 
                            "QLinearAdd", "Pad", "Split", "Gather", "Reshape", "Concat",
                            "QuantizeLinear", "DequantizeLinear", "QLinearAveragePool",
                            "Unsqueeze", "Transpose"
            ]
        else:
            int8_op_list = ["ConvInteger", "MatMulInteger", "QAttention",
                            "DynamicQuantizeLSTM", "Gather", "EmbedLayerNormalization",
                            "DynamicQuantizeLinear"
            ]

        for node in model.model.graph.node:
            possible_int8_res = [name for name in int8_op_list if node.op_type.find(name) != -1]
 
            if any(possible_int8_res):
                if self.backend in ["qlinearops", "qdq", "qoperator"]:
                    if node.op_type == "QuantizeLinear" or node.op_type == "DequantizeLinear" \
                            or node.op_type == "DynamicQuantizeLinear":
                        origin_op_type = node.op_type
                    else:
                        origin_op_type = possible_int8_res[0].split('QLinear')[-1]
                else:
                    origin_op_type = possible_int8_res[0].split('Integer')[0]

                if node.op_type in ["Pad", "Split", "Gather", "Concat", "Reshape", "Unsqueeze", 
                    "Squeeze", "Transpose"]:
                    if any([output.endswith('_quantized') for output in node.output]) or \
                        any(['_DequantizeLinear' in inp for inp in node.input]):
                        origin_op_type = node.op_type
                    else:
                        if node.op_type in res:
                            res[node.op_type]['FP32'] += 1
                        continue

                if origin_op_type == "QAttention":
                    origin_op_type = "Attention"
                if origin_op_type == "DynamicQuantizeLSTM":
                    origin_op_type = "LSTM"
                res[origin_op_type]['INT8'] += 1

            elif node.op_type in fp32_op_list and \
                any(['_DequantizeLinear' in inp for inp in node.input]):
                res[node.op_type]['INT8'] += 1

            elif node.op_type in fp32_op_list and node.name in self.quantize_config:
                if self.quantize_config[node.name] not in self.query_handler.get_fallback_list():
                    res[node.op_type]['FP32'] += 1
                else:
                    res[node.op_type][self.quantize_config[node.name].upper()] += 1

        output_data = [[op_type, sum(res[op_type].values()), res[op_type]['INT8'],
            res[op_type]['BF16'], res[op_type]['FP16'], res[op_type]['FP32']] for \
            op_type in res.keys()]
        Statistics(output_data, 
                   header='Mixed Precision Statistics',
                   field_names=["Op Type", "Total", "INT8", "BF16", "FP16", "FP32"]).print_stat()

    def _get_quantize_params(self, model, data_loader, quantize_config, iterations):
        from neural_compressor.adaptor.ox_utils.onnxrt_mid import ONNXRTAugment
        from neural_compressor.model.onnx_model import ONNXModel
        if not isinstance(model, ONNXModel):
            model = ONNXModel(model)
        black_nodes = [node for node in quantize_config if quantize_config[node]=='fp32']
        white_nodes = [node for node in quantize_config if quantize_config[node]!='fp32']
        augment = ONNXRTAugment(model, \
                  data_loader, self.quantizable_op_types, \
                  os.path.join(self.work_space, 'augmented_model.onnx'), \
                  black_nodes=black_nodes, white_nodes=white_nodes, \
                  iterations=list(range(0, quantize_config['calib_iteration'])))
        self.min_max = augment.dump_minmax()
        quantize_params = augment.dump_calibration()
        return quantize_params

    def inspect_tensor(self, model, dataloader, op_list=[],
                       iteration_list=[],
                       inspect_type='activation',
                       save_to_disk=False,
                       save_path=None,
                       quantization_cfg=None):
        '''The function is used by tune strategy class for dumping tensor info.
        '''
        from neural_compressor.adaptor.ox_utils.onnxrt_mid import ONNXRTAugment
        from neural_compressor.model.onnx_model import ONNXModel
        from neural_compressor.utils.utility import dump_data_to_local
        if not isinstance(model, ONNXModel):
            model = ONNXModel(model)

        if len(op_list) > 0 and isinstance(op_list, KeysView):
            op_list = [item[0] for item in op_list]
        augment = ONNXRTAugment(model, dataloader, [], \
                  os.path.join(self.work_space, 'augment_for_inspect.onnx'), \
                  iterations=iteration_list,
                  white_nodes=op_list)
        tensors = augment.dump_tensor(activation=(inspect_type!='weight'),
                                      weight=(inspect_type!='activation'))
        if save_to_disk:
            if not save_path:
                save_path = self.work_space
            dump_data_to_local(tensors, save_path, 'inspect_result.pkl')
        return tensors

    def set_tensor(self, model, tensor_dict):
        from onnx import numpy_helper
        from neural_compressor.model.onnx_model import ONNXModel
        from neural_compressor.adaptor.ox_utils.util import quantize_data_with_scale_zero
        from neural_compressor.adaptor.ox_utils.util import quantize_data_per_channel
        if not isinstance(model, ONNXModel):
            model = ONNXModel(model)
        assert "QuantizeLinear" in [node.op_type for node in model.model.graph.node], \
                                           'adaptor.set_tensor only accept int8 model'
        input_name_to_nodes = model.input_name_to_nodes
        for tensor_name, tensor_value in tensor_dict.items():
            if not tensor_name.endswith('_quantized'):
                tensor_name += '_quantized'
            not_filter = False
            scale_tensor, zo_tensor = model.get_scale_zero(tensor_name)
            if scale_tensor is None or zo_tensor is None:
                not_filter = True
            else:
                scale_value = numpy_helper.to_array(scale_tensor)
                zo_value = numpy_helper.to_array(zo_tensor)
            assert len(input_name_to_nodes[tensor_name]) == 1, \
                    'quantized filter weight should be input of only one node'
            node = input_name_to_nodes[tensor_name][0] #TBD only for conv bias
            node_name = node.name.replace('_quant', '')
            assert node_name in self.quantize_config
            q_type = self.quantize_config[node_name]['weight']['dtype']
            if not_filter:
                new_tensor_value = self._requantize_bias(model, tensor_name, tensor_value)
            elif self.quantize_config[node_name]['weight']['granularity'] == 'per_tensor':
                new_tensor_value = quantize_data_with_scale_zero(
                    tensor_value,
                    q_type,
                    self.quantize_config[node_name]['weight']['scheme'],
                    scale_value,
                    zo_value)
            else:
                new_tensor_value = quantize_data_per_channel(
                    tensor_value,
                    q_type,
                    self.quantize_config[node_name]['weight']['scheme'],
                    scale_value,
                    zo_value)
            model.set_initializer(tensor_name, new_tensor_value)
        return model

    def _requantize_bias(self, model, bias_name, bias_data):
        ''' helper function to requantize bias, borrowed from onnx_quantizer '''
        from onnx import numpy_helper
        node = model.input_name_to_nodes[bias_name][0]
        input_scale_name = node.input[1]
        input_scale = numpy_helper.to_array(model.get_initializer(input_scale_name))

        weight_scale_name = node.input[4]
        weight_scale = numpy_helper.to_array(model.get_initializer(weight_scale_name))

        bias_scale = input_scale * weight_scale
        new_bias_data = (bias_data / bias_scale).round().astype(np.int32)
        return new_bias_data

    def _pre_optimize(self, model, level=1):
        from neural_compressor.adaptor.ox_utils.util import \
            remove_init_from_model_input, split_shared_bias
        remove_init_from_model_input(model)
        sess_options = ort.SessionOptions()
        level = self.query_handler.get_graph_optimization()
        if self.graph_optimization.level:
            optimization_levels = {
                    'DISABLE_ALL': ort.GraphOptimizationLevel.ORT_DISABLE_ALL,
                    'ENABLE_BASIC': ort.GraphOptimizationLevel.ORT_ENABLE_BASIC,
                    'ENABLE_EXTENDED': ort.GraphOptimizationLevel.ORT_ENABLE_EXTENDED,
                    'ENABLE_ALL': ort.GraphOptimizationLevel.ORT_ENABLE_ALL}
            assert self.graph_optimization.level in optimization_levels, "the optimization \
                                      choices are {}".format(optimization_levels.keys())
 
            level = optimization_levels[self.graph_optimization.level]
        sess_options.graph_optimization_level = level
        sess_options.optimized_model_filepath = os.path.join(self.work_space, \
            "Optimized_model.onnx")
        sess_options.register_custom_ops_library(ort_ext.get_library_path())
        _ = ort.InferenceSession(model.model.SerializeToString(), sess_options)
        tmp_model = onnx.load(sess_options.optimized_model_filepath)
        model.model = self._replace_gemm_with_matmul(tmp_model).model \
            if self.graph_optimization.gemm2matmul else tmp_model
        model.model = self._rename_node(model.model)
        model = self._revert_fusedconv(model)
        model = split_shared_bias(model)
        model.topological_sort()
        self.pre_optimized_model = model

    def _revert_fusedconv(self, model):
        from onnxruntime.quantization.quant_utils import attribute_to_kwarg
        from onnx import onnx_pb as onnx_proto
        new_nodes = []
        remove_nodes = []
        for node in model.model.graph.node:
            if node.op_type == 'FusedConv':
                kwargs = {}
                activation_params = None
                for attr in node.attribute:
                    if attr.name == 'activation':
                        activation_type = attr.s.decode('utf-8')
                    elif attr.name == 'activation_params':
                        continue
                    else:
                        kwargs.update(attribute_to_kwarg(attr))
                if activation_type in ['Relu', 'Clip']:
                    continue
                conv = onnx.helper.make_node(
                    'Conv', node.input, [node.name], node.name.split('fused ')[-1], **kwargs)
                activation_input = conv.output

                activation = onnx.helper.make_node(activation_type,
                    conv.output, node.output, '_'.join((conv.name, activation_type)))
                new_nodes.extend([conv, activation])
                remove_nodes.append(node)
        model.model.graph.node.extend(new_nodes)
        for node in remove_nodes:
            model.model.graph.node.remove(node)
        model.update()
        return model

    def _rename_node(self, model):
        node_names = [i.name for i in model.graph.node]
        if len(set(node_names)) < len(node_names):
            logger.warning("This model has nodes with the same name, please check \
                renamed_model.onnx in workspace_path (default is nc_workspace) \
                for newly generated node name")
            for idx, node in enumerate(model.graph.node):
                if node_names.count(node.name) > 1:
                    node.name = node.op_type + '_nc_rename_' + str(idx)
            onnx.save(model, os.path.join(self.work_space, "renamed_model.onnx")) 
        return model

    @staticmethod
    def _replace_gemm_with_matmul(model):
        new_nodes = []
        from onnx import numpy_helper
        from neural_compressor.model.onnx_model import ONNXModel
        if not isinstance(model, ONNXModel):
            model = ONNXModel(model)

        for node in model.nodes():
            if node.op_type == 'Gemm':
                alpha = 1.0
                beta = 1.0
                transA = 0
                transB = 0
                for attr in node.attribute:
                    if attr.name == 'alpha':
                        alpha = onnx.helper.get_attribute_value(attr)
                    elif attr.name == 'beta':
                        beta = onnx.helper.get_attribute_value(attr)
                    elif attr.name == 'transA':
                        transA = onnx.helper.get_attribute_value(attr)
                    elif attr.name == 'transB':
                        transB = onnx.helper.get_attribute_value(attr)
                if alpha == 1.0 and beta == 1.0 and transA == 0:
                    inputB = node.input[1]
                    if transB == 1:
                        B = model.get_initializer(node.input[1])
                        if B:
                            # assume B is not used by any other node
                            B_array = numpy_helper.to_array(B)
                            B_trans = numpy_helper.from_array(B_array.T)
                            B_trans.name = B.name
                            model.remove_initializer(B)
                            model.add_initializer(B_trans)

                            #TBD this is for onnx model zoo, which are all in old IR version
                            if model.model.ir_version < 4:
                                for input in model.model.graph.input:
                                    if input.name == B_trans.name:
                                        for i, dim in enumerate(input.type.tensor_type.shape.dim):
                                            dim.dim_value = B_array.T.shape[i]

                        else:
                            inputB += '_Transposed'
                            transpose_node = onnx.helper.make_node('Transpose',
                                                                inputs=[node.input[1]],
                                                                outputs=[inputB],
                                                                name=node.name+'_Transpose')
                            new_nodes.append(transpose_node)

                    matmul_node = onnx.helper.make_node('MatMul',
                            inputs=[node.input[0], inputB],
                            outputs=[node.output[0] + ('_MatMul' if len(node.input)>2 else '')],
                            name=node.name + '_MatMul')
                    new_nodes.append(matmul_node)

                    if len(node.input) > 2:
                        add_node = onnx.helper.make_node('Add',
                            inputs=[node.output[0] + '_MatMul', node.input[2]],
                            outputs=node.output,
                            name=node.name + '_Add')
                        new_nodes.append(add_node)

                # unsupported
                else:
                    new_nodes.append(node)

            # not GEMM
            else:
                new_nodes.append(node)

        model.graph().ClearField('node')
        model.graph().node.extend(new_nodes)

        return model

    def query_fw_capability(self, model):
        """The function is used to query framework capability.
        TODO: will be replaced by framework query API

        Args:
            model: onnx model

        Returns:
            (dict): quantization capability
        """
        # optype_wise and op_wise capability
        self._pre_optimize(model)
        exclude_first_quantizable_op = True if 'first_conv_or_matmul_quantization' in \
            self.recipes and not self.recipes['first_conv_or_matmul_quantization'] \
            else False
 
        precisions = self.query_handler.get_precisions()
        optype_wise = OrderedDict()
        quantizable_optype = set([i.op_type for i in self.pre_optimized_model.nodes()])

        op_wise = OrderedDict()
        for precision in precisions:
            if precision == 'fp16' and self.device == 'cpu' and os.getenv('FORCE_FP16') != '1':
                continue
            if precision in self.query_handler.get_quantization_capability():
                special_config_types = list(self.query_handler.get_quantization_capability() \
                    [precision].keys())
                default_config = self.query_handler.get_quantization_capability() \
                    [precision]['default']
            else:
                special_config_types = {}
                default_config = {'weight': {'dtype': [precision]}, 
                                  'activation': {'dtype': [precision]}}
            optypes = self.query_handler.get_op_types_by_precision(precision) if \
                self.query_handler.get_op_types_by_precision(precision) != ['*'] else \
                optype_wise.keys()
            for op in optypes:
                if op not in quantizable_optype:
                    continue
                if op not in special_config_types:
                    op_capability = default_config
                else:
                    op_capability = self.query_handler.get_quantization_capability() \
                        [precision][op]
                if op not in optype_wise.keys():
                    optype_wise[op] = copy.deepcopy(op_capability)
                elif precision not in optype_wise[op]['weight']['dtype']:
                    optype_wise[op]['weight']['dtype'].append(precision)
                    optype_wise[op]['activation']['dtype'].append(precision)

        for key, val in optype_wise.items():
            optype_wise[key]['weight']['dtype'] = \
                [i for i in precisions if i in val['weight']['dtype']]
            optype_wise[key]['activation']['dtype'] = \
                [i for i in precisions if i in val['activation']['dtype']]

        for _, node in enumerate(self.pre_optimized_model.nodes()):
            if node.op_type in optype_wise:
                if exclude_first_quantizable_op and node.op_type in ['Conv', 'MatMul']:
                    exclude_first_quantizable_op = False
                    tmp_cfg = copy.deepcopy(optype_wise[node.op_type])
                    for k, v in tmp_cfg.items():
                        v['dtype'] = list(filter(lambda x: x not in ['uint8', 'int8'], v['dtype']))
                    op_wise.update({(node.name, node.op_type): tmp_cfg})
                    continue
                op_wise.update(
                    {(node.name, node.op_type): copy.deepcopy(optype_wise[node.op_type])})
        return {'optypewise': optype_wise, 'opwise': op_wise}

    def _cfg_to_quantize_config(self, tune_cfg):
        quantize_config = {}
        quantize_config['calib_iteration'] = tune_cfg['calib_iteration']
        granularity = 'per_tensor'
        algorithm = 'minmax'

        from onnx import onnx_pb as onnx_proto
        for _, op in enumerate(self.quantizable_ops):
            if tune_cfg['op'][(op.name, op.op_type)]['activation']['dtype'] in \
                self.query_handler.get_fallback_list():
                quantize_config[op.name] = \
                    tune_cfg['op'][(op.name, op.op_type)]['activation']['dtype']
            else:
                node_config = copy.deepcopy(tune_cfg['op'][(op.name, op.op_type)])
                for tensor, config in tune_cfg['op'][(op.name, op.op_type)].items():
                    if 'granularity' not in config:
                        node_config[tensor]['granularity'] = granularity
                    if 'algorithm' not in config:
                        node_config[tensor]['algorithm'] = algorithm
                    if config['dtype'] == "int8":
                        node_config[tensor]['dtype'] = onnx_proto.TensorProto.INT8
                        if 'scheme' not in config:
                            node_config[tensor]['scheme'] = 'sym'
                    else:
                        node_config[tensor]['dtype'] = onnx_proto.TensorProto.UINT8
                        if 'scheme' not in config:
                            node_config[tensor]['scheme'] = 'asym'
                quantize_config[op.name] = node_config

        return quantize_config

    def _query_quantizable_ops(self, model):
        for node in model.graph.node:
            if node.op_type in self.quantizable_op_types and node not in self.quantizable_ops:
                self.quantizable_ops.append(node)

        return self.quantizable_ops

    def _query_quantizable_op_types(self):
        quantizable_op_types = self.query_handler.get_op_types_by_precision(precision='int8')
        return quantizable_op_types

    def evaluate(self, input_graph, dataloader, postprocess=None,
                 metrics=None, measurer=None, iteration=-1,
                 tensorboard=False, fp32_baseline=False):
        """The function is for evaluation if no given eval func

        Args:
            input_graph      : onnx model for evaluation
            dataloader       : dataloader for evaluation. neural_compressor.data.dataloader.ONNXDataLoader
            postprocess      : post-process for evalution. neural_compressor.data.transform.ONNXTransforms
            metrics:         : metrics for evaluation. neural_compressor.metric.ONNXMetrics
            measurer         : neural_compressor.objective.Measurer
            iteration(int)   : max iterations of evaluaton.
            tensorboard(bool): whether to use tensorboard for visualizaton
            fp32_baseline (boolen, optional): only for compare_label=False pipeline

        Returns:
            (float) evaluation results. acc, f1 e.g.
        """
        sess_options = ort.SessionOptions()
        if measurer:
            # https://github.com/microsoft/onnxruntime/issues/7347
            cores_per_instance = int(os.environ.get('CORES_PER_INSTANCE'))
            assert cores_per_instance > 0, "benchmark cores_per_instance should greater than 0"
            sess_options.intra_op_num_threads = cores_per_instance
        try:
            session = ort.InferenceSession(input_graph.model.SerializeToString(), sess_options)
        except:
            sess_options.register_custom_ops_library(ort_ext.get_library_path())
            session = ort.InferenceSession(input_graph.model.SerializeToString(), sess_options)
        results = []
        if metrics:
            for metric in metrics:
                metric.reset()
            self.fp32_preds_as_label = any([hasattr(metric, "compare_label") and \
                not metric.compare_label for metric in metrics]) 

        ort_inputs = {}
        len_inputs = len(session.get_inputs())
        inputs_names = [session.get_inputs()[i].name for i in range(len_inputs)]

        def eval_func(dataloader):
            for idx, (inputs, labels) in enumerate(dataloader):
                if not isinstance(labels, list):
                    labels = [labels]
                if len_inputs == 1:
                    ort_inputs.update(
                        inputs if isinstance(inputs, dict) else {inputs_names[0]: inputs}
                    )
                else:
                    assert len_inputs == len(inputs), \
                        'number of input tensors must align with graph inputs'

                    if isinstance(inputs, dict):  # pragma: no cover
                        ort_inputs.update(inputs)
                    else:
                        for i in range(len_inputs):
                            # in case dataloader contains non-array input
                            if not isinstance(inputs[i], np.ndarray):
                                ort_inputs.update({inputs_names[i]: np.array(inputs[i])})
                            else:
                                ort_inputs.update({inputs_names[i]: inputs[i]})

                if measurer is not None:
                    measurer.start()
                    predictions = session.run(None, ort_inputs)
                    measurer.end()
                else:
                    predictions = session.run(None, ort_inputs)

                if self.fp32_preds_as_label:
                    self.fp32_results.append(predictions) if fp32_baseline else \
                        results.append(predictions)

                if postprocess is not None:
                    predictions, labels = postprocess((predictions, labels))
                if metrics:
                    for metric in metrics:
                        if not hasattr(metric, "compare_label") or \
                            (hasattr(metric, "compare_label") and metric.compare_label):
                            metric.update(predictions, labels)
                if idx + 1 == iteration:
                    break

        if isinstance(dataloader, BaseDataLoader) and not self.benchmark:
            try:
                eval_func(dataloader)
            except Exception:  # pragma: no cover
                logger.warning(
                    "Fail to forward with batch size={}, set to {} now.".
                    format(dataloader.batch_size, 1))
                dataloader.batch(1)
                eval_func(dataloader)
        else:  # pragma: no cover
            eval_func(dataloader)

        if self.fp32_preds_as_label:
            from neural_compressor.adaptor.ox_utils.util import collate_preds
            if fp32_baseline:
                results = collate_preds(self.fp32_results)
                reference = results
            else:
                reference = collate_preds(self.fp32_results)
                results = collate_preds(results)
            for metric in metrics:
                if hasattr(metric, "compare_label") and not metric.compare_label:
                    metric.update(results, reference)

        acc = 0 if metrics is None else [metric.result() for metric in metrics]
        return acc if not isinstance(acc, list) or len(acc) > 1 else acc[0]

    def diagnosis_helper(self, fp32_model, int8_model, tune_cfg=None, save_path=None):
        from neural_compressor.utils.utility import dump_data_to_local
        from neural_compressor.model.onnx_model import find_by_name
        if self.backend in ["qlinearops", "qoperator"]:
            supported_optype = ['Conv', 'MatMul', 'Concat', 'Attention', 'FusedConv',
                'Add', 'Mul', 'LeakyRelu', 'Sigmoid', 'GlobalAveragePool', 'AveragePool']
        elif self.backend == "qdq":
            supported_optype = ['Conv', 'MatMul', 'Concat', 'Attention', 'FusedConv',
                'LeakyRelu', 'Sigmoid', 'GlobalAveragePool', 'AveragePool']
        else:
            supported_optype = ['Conv', 'MatMul', 'Attention', 'LSTM']
        inspect_node_list = []
        int8_node_names = [i.name for i in int8_model.nodes()]
        for node in fp32_model.nodes():
            if node.op_type in supported_optype and node.name + '_quant' in int8_node_names:
                inspect_node_list.append(node.name)

        filtered_params = {}
        if self.min_max:
            for node_name in inspect_node_list:
                node = find_by_name(node_name, fp32_model.nodes())
                filtered_params[node_name] = {
                    'min': np.array(self.min_max[node.output[0]][0], dtype=np.float32),
                    'max': np.array(self.min_max[node.output[0]][1], dtype=np.float32)}
        if save_path:
            dump_data_to_local(filtered_params, save_path, 'dequan_min_max.pkl')
            dump_data_to_local(tune_cfg, save_path, 'cfg.pkl')
        return inspect_node_list, tune_cfg

    def save(self, model, path):
        """ save model

        Args:
            model (ModelProto): model to save
            path (str): save path
        """
        model.save(os.path.join(path, "best_model.onnx"))


@adaptor_registry
class ONNXRT_QLinearOpsAdaptor(ONNXRTAdaptor):
    """The ONNXRT adaptor layer, do onnx-rt quantization, calibration, inspect layer tensors.

    Args:
        framework_specific_info (dict): framework specific configuration for quantization.
    """

    def __init__(self, framework_specific_info):
        self.query_handler = ONNXRTQuery(local_config_file=os.path.join(
            os.path.dirname(__file__), "onnxrt_qlinear.yaml"))
        self.backend = "qlinearops"
        super().__init__(framework_specific_info)

@adaptor_registry
class ONNXRT_QOperatorAdaptor(ONNXRTAdaptor):
    """The ONNXRT adaptor layer, do onnx-rt quantization, calibration, inspect layer tensors.

    Args:
        framework_specific_info (dict): framework specific configuration for quantization.
    """

    def __init__(self, framework_specific_info):
        self.query_handler = ONNXRTQuery(local_config_file=os.path.join(
            os.path.dirname(__file__), "onnxrt_qlinear.yaml"))
        self.backend = "qlinearops"
        super().__init__(framework_specific_info)


@adaptor_registry
class ONNXRT_IntegerOpsAdaptor(ONNXRTAdaptor):
    """The ONNXRT adaptor layer, do onnx-rt quantization, calibration, inspect layer tensors.

    Args:
        framework_specific_info (dict): framework specific configuration for quantization.
    """

    def __init__(self, framework_specific_info):
        self.query_handler = ONNXRTQuery(local_config_file=os.path.join(
            os.path.dirname(__file__), "onnxrt_integer.yaml"))
        self.backend = "integerops"
        super().__init__(framework_specific_info)

@adaptor_registry
class ONNXRT_QDQAdaptor(ONNXRTAdaptor):
    """The ONNXRT adaptor layer, do onnx-rt quantization, calibration, inspect layer tensors.

    Args:
        framework_specific_info (dict): framework specific configuration for quantization.
    """

    def __init__(self, framework_specific_info):
        self.query_handler = ONNXRTQuery(local_config_file=os.path.join(
            os.path.dirname(__file__), "onnxrt_qdq.yaml"))
        self.backend = "qdq"
        super().__init__(framework_specific_info)

class ONNXRTQuery(QueryBackendCapability):

    def __init__(self, local_config_file=None):
        super().__init__()
        self.version = ort.__version__
        self.cfg = local_config_file
        self.cur_config = None
        self._one_shot_query()

    def _one_shot_query(self):
        with open(self.cfg) as f:
            content = yaml.safe_load(f)
            try:
                self.cur_config = self._get_specified_version_cfg(content)
            except Exception as e: # pragma: no cover
                logger.info("Fail to parse {} due to {}.".format(self.cfg, str(e)))
                self.cur_config = None
                raise ValueError("Please check if the format of {} follows Neural Compressor yaml schema.".
                                 format(self.cfg))

    def _get_specified_version_cfg(self, data): # pragma: no cover
        """Get the configuration for the current runtime.
        If there's no matched configuration in the input yaml, we'll
        use the `default` field of yaml.

        Args:
            data (Yaml content): input yaml file.

        Returns:
            [dictionary]: the content for specific version.
        """
        default_config = None
        for sub_data in data:
            if self.version in sub_data['version']['name']:
                return sub_data

            if 'default' in sub_data['version']['name']:
                default_config = sub_data

        return default_config

    def get_version(self): # pragma: no cover
        """Get the current backend version infomation.

        Returns:
            [string]: version string.
        """
        return self.cur_config['version']['name']

    def get_precisions(self): # pragma: no cover
        """Get supported precisions for current backend.

        Returns:
            [string list]: the precisions' name.
        """
        return [i.strip() for i in self.cur_config['precisions']['names'].split(',')]

    def get_op_types(self): # pragma: no cover
        """Get the supported op types by all precisions.

        Returns:
            [dictionary list]: A list composed of dictionary which key is precision
            and value is the op types.
        """
        return self.cur_config['ops']

    def get_quantization_capability(self):
        """Get the supported op types' quantization capability.

        Returns:
            [dictionary list]: A list composed of dictionary which key is precision
            and value is a dict that describes all op types' quantization capability.
        """
        return self.cur_config['capabilities']

    def get_op_types_by_precision(self, precision):
        """Get op types per precision

        Args:
            precision (string): precision name

        Returns:
            [string list]: A list composed of op type.
        """
        #assert precision in list(self.cur_config['ops'].keys())
        if precision in list(self.cur_config['ops'].keys()):
            return self.cur_config['ops'][precision]
        else:
            return []

    def get_graph_optimization(self):
        """ Get onnxruntime graph optimization level"""
        optimization_levels = {'DISABLE_ALL': ort.GraphOptimizationLevel.ORT_DISABLE_ALL,
                               'ENABLE_BASIC': ort.GraphOptimizationLevel.ORT_ENABLE_BASIC,
                               'ENABLE_EXTENDED': ort.GraphOptimizationLevel.ORT_ENABLE_EXTENDED,
                               'ENABLE_ALL': ort.GraphOptimizationLevel.ORT_ENABLE_ALL}
 
        level = self.cur_config['graph_optimization']['level']
        assert level in optimization_levels, "the optimization choices \
                                              are {}".format(optimization_levels.keys())
        return optimization_levels[level]

    def get_fallback_list(self):
        return list(self.cur_config['ops'].keys() - self.cur_config['capabilities'].keys())



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# whethre to consider all import modules and perform transformation based on all codes plus all import modules
import logging
consider_imports = True

# target batch size for feature of changing PyTorch batch size
target_batch_size = 1

# number of benchmark iteration for feature of PyTorch benchmark
num_benchmark_iteration = 30

# print info for debugging purpose
logging_level = logging.INFO


def reset_globals():
    global list_code_path

    global list_code_line_instance

    global list_class_def_instance
    global list_class_name
    global list_parent_class_name

    global list_model_def_instance
    global list_model_name

    global list_trans_insert_modified_file
    global list_trans_insert_location_idxs
    global list_trans_insert_number_insert_lines
    global list_trans_insert_lines_to_insert

    global list_trans_indenting_modified_file
    global list_trans_indenting_location_idxs
    global list_trans_indenting_level

    global list_all_function_name
    global list_all_function_return_item

    global list_wrapper_base_function_name
    global list_wrapper_children_function_name
    global list_wrapper_all_function_name

    list_code_path = []
    list_code_line_instance = []  # list of CodeLine instances

    list_class_def_instance = []  # list of ClassDefinition instances
    list_class_name = []  # list of class names
    list_parent_class_name = []

    list_model_def_instance = []
    list_model_name = []

    # for code transformation recording
    list_trans_insert_modified_file = []
    list_trans_insert_location_idxs = []
    list_trans_insert_number_insert_lines = []
    list_trans_insert_lines_to_insert = []

    list_trans_indenting_modified_file = []
    list_trans_indenting_location_idxs = []
    list_trans_indenting_level = []

    list_all_function_name = []
    list_all_function_return_item = []

    list_wrapper_base_function_name = []
    list_wrapper_children_function_name = []
    list_wrapper_all_function_name = []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neural_compressor.utils.utility import dump_elapsed_time

from ..graph_base import GraphRewriterBase
from neural_compressor.adaptor.tf_utils.graph_util import GraphAnalyzer


class MoveSqueezeAfterReluOptimizer(GraphRewriterBase):
    def __init__(self, model):
        super().__init__(model)
        self.op_list = ['Relu', 'Sigmoid']

    @dump_elapsed_time("Pass MoveSqueezeAfterReluOptimizer")
    def do_transformation(self):
        g = GraphAnalyzer()
        g.graph = self.model
        graph_info = g.parse_graph()
        for node in self.model.node:
            if node.op in self.op_list and \
                    node.input[0] in graph_info and \
                        graph_info[node.input[0]].node.op == 'BiasAdd':
                biasadd_node = graph_info[node.input[0]].node
                biasadd_input = graph_info[biasadd_node.name].node.input[0]
                squeeze_node = graph_info[biasadd_input].node
                relu_output = graph_info[node.name].outputs
                if squeeze_node.op == 'Squeeze':
                    #biasadd
                    for i, input in enumerate(biasadd_node.input):
                        if input == biasadd_input:
                            new_input = biasadd_node.input[:i] + [squeeze_node.input[0]] + \
                                    biasadd_node.input[i+1:]
                            graph_info[biasadd_node.name].node.ClearField('input')
                            graph_info[biasadd_node.name].node.input.extend(new_input)
                            graph_info[squeeze_node.name].outputs.remove(biasadd_node.name)
                    #conv output
                    conv = squeeze_node.input[0]
                    conv_outputs = graph_info[conv].outputs
                    for i, output in enumerate(conv_outputs):
                        if output == squeeze_node.name:
                            graph_info[conv].outputs.remove(squeeze_node.name)
                            graph_info[conv].outputs.append(biasadd_node.name)
                    #squeeze input
                    squeeze_node.ClearField('input')
                    squeeze_node.input.extend([node.name])
                    #expand input,squeeze output
                    for output in relu_output:
                        for i, input in enumerate(graph_info[output].node.input):
                            if input == node.name:
                                new_input = graph_info[output].node.input[:i] + [squeeze_node.name] +\
                                graph_info[output].node.input[i+1:]
                                graph_info[squeeze_node.name].outputs.append(output)
                                graph_info[output].node.ClearField('input')
                                graph_info[output].node.input.extend(new_input)
        return g.dump_graph()



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ... import globals
from ...utils.line_operation import get_line_indent_level, is_eval_func_model_name

import logging

logging.basicConfig(level=globals.logging_level,
                    format='%(asctime)s %(levelname)s %(message)s',
                    datefmt='%a, %d %b %Y %H:%M:%S +0000')
logger = logging.getLogger(__name__)


class Benchmark(object):
    def __init__(self):
        pass

    # collect file transformation info and register (store) in globals
    # (i.e. which file to add which lines at which location)
    def register_transformation(self):
        for file_path in globals.list_code_path:
            try:
                code = open(file_path, 'r').read()
                lines = code.split('\n')
                line_idx = 0
                for i in range(len(lines)):
                    line = lines[i]
                    for model_name in globals.list_model_name:
                        if is_eval_func_model_name(model_name, line) and "# Neural Coder appended" not in line:
                            indent_level = get_line_indent_level(line)

                            # 1. indenting
                            # indenting can have multiple location, so is a list of numbers
                            trans_indenting_location = []
                            trans_indenting_level = []

                            if ")" in line:  # e.g. model(xxx)
                                trans_indenting_location.append(line_idx)
                                trans_indenting_level.append(1)
                            else:  # e.g. model(xxx，
                                #            xxx，
                                #            xxx
                                #      )
                                trans_indenting_location.append(line_idx)
                                trans_indenting_level.append(1)
                                do_search = True
                                i_search = 1
                                while do_search:
                                    trans_indenting_location.append(
                                        line_idx + i_search)
                                    trans_indenting_level.append(1)
                                    following_line = lines[line_idx + i_search]
                                    if ")" in following_line:
                                        do_search = False
                                    i_search += 1

                            # 1. register indenting: transform "model(input)" to "    model(input)"
                            if file_path not in globals.list_trans_indenting_modified_file:
                                globals.list_trans_indenting_modified_file.append(
                                    file_path)
                                globals.list_trans_indenting_location_idxs.append(
                                    trans_indenting_location)
                                globals.list_trans_indenting_level.append(
                                    trans_indenting_level)
                            else:
                                idx = globals.list_trans_indenting_modified_file.index(
                                    file_path)
                                for i in trans_indenting_location:
                                    globals.list_trans_indenting_location_idxs[idx].append(
                                        i)
                                for i in trans_indenting_level:
                                    globals.list_trans_indenting_level[idx].append(
                                        i)

                            # 2-1. insert (before model(input))
                            trans_insert_location = []  # insert only has 1 location, so is a number

                            trans_insert_location = line_idx  # insert before

                            lines_to_insert = ""
                            lines_to_insert += " " * indent_level + "import time" + "\n"
                            lines_to_insert += " " * indent_level + "count_iter_ = 0" + "\n"
                            lines_to_insert += " " * indent_level + "total_time_ = 0" + "\n"
                            lines_to_insert += " " * indent_level + "num_iter_ = " + \
                                globals.num_benchmark_iteration + "\n"
                            lines_to_insert += " " * indent_level + "num_warmup_iter_ = 10" + "\n"
                            lines_to_insert += " " * indent_level + "list_batch_time_ = []" + "\n"
                            lines_to_insert += " " * indent_level + \
                                "for i_ in range(num_iter_):" + "\n"
                            lines_to_insert += " " * indent_level + " " * \
                                4 + "count_iter_ = count_iter_ + 1" + "\n"
                            lines_to_insert += " " * indent_level + " " * 4 + \
                                "if count_iter_ > num_warmup_iter_:" + "\n"
                            lines_to_insert += " " * indent_level + " " * 8 + "t1_ = time.time()"

                            # 2-1. register insert (before model(input))
                            if file_path not in globals.list_trans_insert_modified_file:
                                globals.list_trans_insert_modified_file.append(
                                    file_path)
                                globals.list_trans_insert_location_idxs.append(
                                    [trans_insert_location])
                                globals.list_trans_insert_number_insert_lines.append(
                                    [lines_to_insert.count("\n") + 1])
                                globals.list_trans_insert_lines_to_insert.append(
                                    [lines_to_insert])
                            else:
                                idx = globals.list_trans_insert_modified_file.index(
                                    file_path)
                                globals.list_trans_insert_location_idxs[idx].append(
                                    trans_insert_location)
                                globals.list_trans_insert_number_insert_lines[idx].append(
                                    lines_to_insert.count("\n") + 1)
                                globals.list_trans_insert_lines_to_insert[idx].append(
                                    lines_to_insert)

                            # 2-2. insert (after model(input))
                            trans_insert_location = []  # insert only has 1 location, so is a number

                            if ")" in line:  # e.g. model()
                                trans_insert_location = line_idx + 1  # insert after
                            else:  # e.g. model(xxx，
                                #              xxx，
                                #              xxx
                                #        )
                                do_search = True
                                i_search = 1
                                while do_search:
                                    following_line = lines[line_idx + i_search]
                                    if ")" in following_line and following_line[indent_level] == ")":
                                        do_search = False
                                    i_search += 1
                                trans_insert_location = line_idx + i_search  # insert after

                            lines_to_insert = ""
                            lines_to_insert += " " * indent_level + " " * 4 + \
                                "if count_iter_ > num_warmup_iter_:" + "\n"
                            lines_to_insert += " " * indent_level + " " * 8 + "t2_ = time.time()" + "\n"
                            lines_to_insert += " " * indent_level + \
                                " " * 8 + "batch_time_ = t2_ - t1_" + "\n"
                            lines_to_insert += " " * indent_level + " " * 8 + \
                                "list_batch_time_.append(batch_time_)" + "\n"
                            lines_to_insert += " " * indent_level + " " * 8 + \
                                "total_time_ = total_time_ + batch_time_" + "\n"
                            lines_to_insert += " " * indent_level + \
                                'print("Neural_Coder_Bench_IPS: ",' \
                                'round((num_iter_ - num_warmup_iter_) / total_time_, 3))' + "\n"
                            lines_to_insert += " " * indent_level + \
                                'print("Neural_Coder_Bench_MSPI: ",' \
                                'round(total_time_ / (num_iter_ - num_warmup_iter_) * 1000, 3))' + "\n"
                            lines_to_insert += " " * indent_level + "list_batch_time_.sort()" + "\n"
                            lines_to_insert += " " * indent_level + \
                                "p50_latency_ = list_batch_time_[int(len(list_batch_time_) * 0.50) - 1] * 1000" \
                                + "\n"
                            lines_to_insert += " " * indent_level + \
                                "p90_latency_ = list_batch_time_[int(len(list_batch_time_) * 0.90) - 1] * 1000" \
                                + "\n"
                            lines_to_insert += " " * indent_level + \
                                "p99_latency_ = list_batch_time_[int(len(list_batch_time_) * 0.99) - 1] * 1000" \
                                + "\n"
                            lines_to_insert += " " * indent_level + \
                                'print("Neural_Coder_Bench_P50: ", round(p50_latency_, 3))' + "\n"
                            lines_to_insert += " " * indent_level + \
                                'print("Neural_Coder_Bench_P90: ", round(p90_latency_, 3))' + "\n"
                            lines_to_insert += " " * indent_level + \
                                'print("Neural_Coder_Bench_P99: ", round(p99_latency_, 3))' + "\n"

                            # 2-2. register insert (after model(input))
                            if file_path not in globals.list_trans_insert_modified_file:
                                globals.list_trans_insert_modified_file.append(
                                    file_path)
                                globals.list_trans_insert_location_idxs.append(
                                    [trans_insert_location])
                                globals.list_trans_insert_number_insert_lines.append(
                                    [lines_to_insert.count("\n") + 1])
                                globals.list_trans_insert_lines_to_insert.append(
                                    [lines_to_insert])
                            else:
                                idx = globals.list_trans_insert_modified_file.index(
                                    file_path)
                                globals.list_trans_insert_location_idxs[idx].append(
                                    trans_insert_location)
                                globals.list_trans_insert_number_insert_lines[idx].append(
                                    lines_to_insert.count("\n") + 1)
                                globals.list_trans_insert_lines_to_insert[idx].append(
                                    lines_to_insert)

                    line_idx += 1
            except:
                logger.debug(
                    f"This file has skipped patching due to unrecognizable code format: {file_path}")

        logger.debug(
            f"globals.list_trans_indenting_modified_file: {globals.list_trans_indenting_modified_file}")
        logger.debug(
            f"globals.list_trans_indenting_location_idxs: {globals.list_trans_indenting_location_idxs}")
        logger.debug(
            f"globals.list_trans_indenting_level: {globals.list_trans_indenting_level}")

        logger.debug(
            f"globals.list_trans_insert_modified_file: {globals.list_trans_insert_modified_file}")
        logger.debug(
            f"globals.list_trans_insert_location_idxs: {globals.list_trans_insert_location_idxs}")
        logger.debug(
            f"globals.list_trans_insert_number_insert_lines: {globals.list_trans_insert_number_insert_lines}")
        logger.debug(
            f"globals.list_trans_insert_lines_to_insert: {globals.list_trans_insert_lines_to_insert}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import onnx
from .base_operator import QuantOperatorBase
from onnxruntime.quantization.quant_utils import attribute_to_kwarg, ms_domain, \
                                                 QuantizedValueType
from neural_compressor.adaptor.ox_utils.util import QuantizedValue                                                 
class QGlobalAveragePool(QuantOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def convert(self):
        node = self.node
        assert (node.op_type == "GlobalAveragePool")

        if len(self.quantizer.model.get_children(node)) == 0:
            return
        parent = self.quantizer.model.get_parents(node)[0]
        child = self.quantizer.model.get_children(node)[0]

        kwargs = {}
        for attribute in node.attribute:
            kwargs.update(attribute_to_kwarg(attribute))
        kwargs["domain"] = ms_domain
        kwargs["channels_last"] = 0

        inputs = parent.input
        inputs.extend(child.input[1:])

        qnode = onnx.helper.make_node(
            "QLinear" + node.op_type,
            inputs,
            child.output,
            node.name, **kwargs)
        self.quantizer.new_nodes += [qnode]
        self.quantizer.remove_nodes.append(child)
        self.quantizer.remove_nodes.append(parent)
        self.quantizer.remove_nodes.append(node)



#
#  -*- coding: utf-8 -*-
#
#  Copyright (c) 2021 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import copy
import os
import logging
import tensorflow as tf

from tensorflow.python.platform import gfile
from neural_compressor.conf.dotdict import deep_get
from neural_compressor.experimental.common import Model
from .transform_graph.rerange_quantized_concat import RerangeQuantizedConcat
from .transform_graph.bias_correction import BiasCorrection
from .quantize_graph.quantize_graph_for_intel_cpu import QuantizeGraphForIntel
from .quantize_graph_common import QuantizeGraphHelper

from .graph_util import GraphAnalyzer
from .graph_rewriter.generic.remove_training_nodes import RemoveTrainingNodesOptimizer
from .graph_rewriter.generic.strip_unused_nodes import StripUnusedNodesOptimizer
from .graph_rewriter.generic.fold_batch_norm import FoldBatchNormNodesOptimizer
from .graph_rewriter.generic.fuse_pad_with_conv import FusePadWithConv2DOptimizer

from .graph_rewriter.int8.freeze_value_without_calib import FreezeValueWithoutCalibTransformer
from .graph_rewriter.int8.fuse_conv_requantize import FuseConvRequantizeTransformer
from .graph_rewriter.int8.fuse_matmul_requantize import FuseMatMulRequantizeTransformer
from .graph_rewriter.int8.fuse_matmul_requantize import FuseMatMulRequantizeDequantizeTransformer
from .graph_rewriter.int8.scale_propagation import ScaleProPagationTransformer
from .graph_rewriter.bf16.bf16_convert import BF16Convert
from .graph_rewriter.int8.post_quantized_op_cse import PostCseOptimizer
from .graph_rewriter.int8.meta_op_optimizer import MetaInfoChangingMemOpOptimizer
from .graph_rewriter.int8.rnn_convert import QuantizedRNNConverter


TF_SUPPORTED_MAX_VERSION = '2.9.1'
TF_SUPPORTED_MIN_VERSION = '1.14.0'

logger = logging.getLogger()
debug = bool(logger.level == logging.DEBUG)

class GraphConverterWithoutCalib:
    def __init__(self,
                 model,
                 data_loader=None,
                 recover_config=None,
                 new_api=False):
        """Convert graph without calibration.

        :param model: input tensorflow model.
        :param qt_config: quantization configs, including interation and op-wise quant config
        :param fp32_ops: fall back to fp32 dtype op list
        :param bf16_ops: fall back to bf16 dtype op list
        :param data_loader: for calibration phase used dataloader
        :param recover_config: config for recovering tuned model
        """
        # Logger initial
        self.model = model
        #(TODO) does it right to make the internal model format as graph_def
        self.output_tensor_names = self.model.output_tensor_names
        self.input_tensor_names = self.model.input_tensor_names
        # quantize specific config
        self.op_wise_config = recover_config['op_wise_config']
        self.advance_config = deep_get(recover_config, 'advance')
        self.device = recover_config['device'] if 'device' in recover_config else 'cpu'
        self.int8_sequences = recover_config['int8_sequences']
        self.fp32_ops = recover_config['fp32_ops']
        self.bf16_ops = recover_config['bf16_ops']
        self.recipes = recover_config['recipes']
        self.quantized_node_info = []
        self._calibration_data = []
        self._fp32_print_data = []
        self.data_loader = data_loader
        self.recover_config = recover_config
        self._check_tf_version()
        self._check_args()
        self._gen_tmp_filenames()
        self.new_api = new_api          
        self._tmp_graph_def = copy.deepcopy(self.model.graph_def)
    # pylint: disable=no-member
    def _check_tf_version(self):
        is_supported_version = False
        try:
            from tensorflow import python
            if (hasattr(python, "pywrap_tensorflow")
                    and hasattr(python.pywrap_tensorflow, "IsMklEnabled")):# pragma: no cover
                from tensorflow.python.pywrap_tensorflow import IsMklEnabled
            elif hasattr(python.util, "_pywrap_util_port"):
                from tensorflow.python.util._pywrap_util_port import IsMklEnabled
            else:
                from tensorflow.python._pywrap_util_port import IsMklEnabled
            if IsMklEnabled() and (TF_SUPPORTED_MIN_VERSION <= tf.version.VERSION):
                is_supported_version = True
                
            if tf.version.VERSION >= '2.6.0' and os.getenv('TF_ENABLE_ONEDNN_OPTS') == '1':
                is_supported_version = True

            if tf.version.VERSION >= '2.9.0':
                is_supported_version = True

        except Exception as e:
            raise ValueError(e)
        finally:# pragma: no cover
            if tf.version.VERSION > TF_SUPPORTED_MAX_VERSION:
                logger.warning(
                    str('Please note the {} version of Intel® Optimizations for '
                        'TensorFlow is not fully verified! '
                        'Suggest to use the versions '
                        'between {} and {} if meet problem.').format(tf.version.VERSION,
                                                                     TF_SUPPORTED_MIN_VERSION,
                                                                     TF_SUPPORTED_MAX_VERSION))
            if tf.version.VERSION == '2.5.0' and os.getenv('TF_ENABLE_MKL_NATIVE_FORMAT') != '0':
                logger.fatal("Please set environment variable TF_ENABLE_MKL_NATIVE_FORMAT=0 "
                             "when TensorFlow 2.5.0 installed.")

            if tf.version.VERSION >= '2.6.0' and tf.version.VERSION < '2.9.0' \
                    and os.getenv('TF_ENABLE_ONEDNN_OPTS') != '1':
                logger.fatal("Please set environment variable TF_ENABLE_ONEDNN_OPTS=1 "
                             "when TensorFlow >= 2.6.0 and < 2.9.0 installed.")

            if not is_supported_version:
                raise ValueError(
                    str('Please install Intel® Optimizations for TensorFlow '
                        'or MKL enabled source build TensorFlow '
                        'within version >={} and <={}.').format(TF_SUPPORTED_MIN_VERSION,
                                                                TF_SUPPORTED_MAX_VERSION))

    def _check_args(self):
        if self.model.workspace_path and not os.path.isdir(self.model.workspace_path) \
                and not os.path.exists(os.path.dirname(self.model.workspace_path)):
            raise ValueError('"output_graph" directory does not exist.')
        self._output_path = self.model.workspace_path

    def _gen_tmp_filenames(self):
        self._int8_dynamic_range_model_path = os.path.join(self._output_path, \
                                                      'int8_dynamic_range_graph')
        self._int8_logged_model_path = os.path.join(self._output_path, 'int8_logged_graph')
        self._fp32_logged_model_path = os.path.join(self._output_path, 'fp32_logged_graph')
        self._int8_frozen_range_model_path = os.path.join(self._output_path,
                                                          'int8_frozen_range_graph')
        self._bf16_mixed_precision_model_path = os.path.join(self._output_path,
                                                        'int8_bf16_mixed_precision_graph')

        self.output_graph = os.path.join(self._output_path, 'int8_final_fused_graph')
        # to keep temp model
        self._tmp_model = Model(self.model._model, **self.model.kwargs)
        self._tmp_model.output_tensor_names = self.output_tensor_names
        self._tmp_model.input_tensor_names = self.input_tensor_names

    def convert_without_calib(self):
        model = self._tmp_model

        if len(self.op_wise_config) > 0:
            model = self.quantize_without_calib()

        if len(self.bf16_ops) > 0:
            model = self.bf16_convert()

        post_cse_graph_def = PostCseOptimizer(model.graph_def).do_transformation()
        post_cse_graph_def.library.CopyFrom(self.model.graph_def.library)
        model.graph_def = post_cse_graph_def

        if debug:
            model.save(self.output_graph)

        return model

    def _analysis_rnn_model(self):
        g = GraphAnalyzer()
        g.graph = self._tmp_graph_def
        graph_info = g.parse_graph()
        rnn_pattern = [['TensorArrayV3'], ['Enter'], ['TensorArrayReadV3'], \
            ['MatMul'], ['BiasAdd']]
        target_nodes = g.query_fusion_pattern_nodes(rnn_pattern)
        res = {}
        for i in target_nodes:
            if i[-3] not in self.bf16_ops and i[-3] not in self.fp32_ops:
                res[(i[-3], i[-2])] = graph_info[i[1]].node.attr['frame_name'].s.decode()

        return res

    def quantize_without_calib(self):
        """Quantize graph only (without optimizing fp32 graph), including:
            1) quantize graph,
            2) fuse RequantizeOp with fused quantized conv, and so on.

        :return:
        """
        try:
            self._quantize_graph()
            self._rnn_details = self._analysis_rnn_model()
            self._freeze_requantization_ranges_without_calib()
            self._fuse_requantize_with_fused_quantized_node()
        except Exception as e:
            import traceback
            traceback.print_exc()
            self._tmp_model = None
            logger.error('Fail to quantize graph due to {}.'.format(str(e)))
        finally:
            if not debug:
                self._post_clean()
            return self._tmp_model

    def bf16_convert(self):
        """Convert fp32 nodes in bf16_node to bf16 dtype based on
           FP32 + INT8 mixed precision graph.
        """
        try:
            self._tmp_model.graph_def = BF16Convert(
                self._tmp_model.graph_def,
                self.fp32_ops,
                self.bf16_ops).do_transformation()

        except Exception as e:
            self._tmp_model = None
            logger.error('Fail to convert graph due to {}.'.format(str(e)))
        finally:
            if debug:
                self._tmp_model.save(self._bf16_mixed_precision_model_path)

            return self._tmp_model

    def _quantize_graph(self):
        """quantize graph."""

        non_pad_ops = list(list(set(self.fp32_ops).union(set(self.bf16_ops))))
        self._tmp_graph_def = FusePadWithConv2DOptimizer(
            self._tmp_graph_def,
            non_pad_ops,
            self._tmp_model.input_node_names,
            self.op_wise_config,
            self.new_api).do_transformation()

        self._tmp_graph_def = QuantizeGraphHelper().get_sorted_graph(
            self._tmp_graph_def,
            self._tmp_model.input_node_names,
            self._tmp_model.output_node_names)

        self._tmp_graph_def, self.quantized_node_info = QuantizeGraphForIntel(
            self._tmp_graph_def,
            self._tmp_model.input_node_names,
            self._tmp_model.output_node_names,
            self.op_wise_config,
            self.int8_sequences,
            self.device,
            False,
            self.new_api).do_transform()

        self._tmp_graph_def.library.CopyFrom(self.model.graph_def.library)
        if debug:
            self._tmp_model.graph_def = self._tmp_graph_def
            self._tmp_model.save(self._int8_dynamic_range_model_path)

    def _freeze_requantization_ranges_without_calib(self):

        self._tmp_graph_def = FreezeValueWithoutCalibTransformer(
            self._tmp_graph_def,
            self.recover_config,
            postfix='__min').do_transformation_without_calib()
        self._tmp_graph_def = FreezeValueWithoutCalibTransformer(
            self._tmp_graph_def,
            self.recover_config,
            postfix='__max').do_transformation_without_calib()
        self._tmp_graph_def = FreezeValueWithoutCalibTransformer(
            self._tmp_graph_def,
            self.recover_config,
            postfix='__requant_min_max',
            device = self.device).do_transformation_without_calib()

        self._tmp_graph_def = QuantizedRNNConverter(
            self._tmp_graph_def, self._calibration_data, self._rnn_details).do_transformation()

        if 'scale_propagation_max_pooling' in self.recipes and \
                self.recipes['scale_propagation_max_pooling']:
            self._tmp_graph_def = ScaleProPagationTransformer(
                self._tmp_graph_def).do_transformation()

        if debug:
            self._tmp_graph_def.library.CopyFrom(self.model.graph_def.library)
            self._tmp_model.graph_def = self._tmp_graph_def
            self._tmp_model.save(self._int8_frozen_range_model_path)

    def _fuse_requantize_with_fused_quantized_node(self):
        self._tmp_graph_def = FuseConvRequantizeTransformer(
            self._tmp_graph_def,
            self.device,
            self.new_api).do_transformation()

        self._tmp_graph_def = FuseMatMulRequantizeTransformer(
            self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = FuseMatMulRequantizeDequantizeTransformer(
            self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = StripUnusedNodesOptimizer(
            self._tmp_graph_def,
            self._tmp_model.input_node_names,
            self._tmp_model.output_node_names).do_transformation()

        self._tmp_graph_def = RemoveTrainingNodesOptimizer(
            self._tmp_graph_def,
            protected_nodes=self._tmp_model.output_node_names).do_transformation()

        self._tmp_graph_def = FoldBatchNormNodesOptimizer(
            self._tmp_graph_def).do_transformation()

        if 'scale_propagation_concat' in self.recipes and self.recipes['scale_propagation_concat']:
            self._tmp_graph_def = RerangeQuantizedConcat(self._tmp_graph_def,
                                                     self.device).do_transformation()

        self._tmp_graph_def = MetaInfoChangingMemOpOptimizer(
            self._tmp_graph_def).do_transformation()

        if self.advance_config is not None and \
           deep_get(self.advance_config, 'bias_correction') is not None:
            self._tmp_graph_def = BiasCorrection(
                self._tmp_graph_def, self.model.graph_def).do_transformation()

        self._tmp_graph_def.library.CopyFrom(self.model.graph_def.library)

        self._tmp_model.graph_def = self._tmp_graph_def

    def _post_clean(self):
        """Delete the temporarily files generated during the quantization process.

        :return: None
        """
        if os.path.exists(self._int8_logged_model_path) and \
            os.path.isdir(self._int8_logged_model_path):
            import shutil
            shutil.rmtree(self._int8_logged_model_path)

        elif gfile.Exists(self._int8_logged_model_path + '.pb'):
            os.remove(self._int8_logged_model_path + '.pb')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neural_compressor.utils.utility import LazyImport
from .base_dataloader import BaseDataLoader
from .default_dataloader import DefaultDataLoader
from ..datasets.bert_dataset import ONNXRTBertDataset
import logging
torch = LazyImport('torch')

class ONNXRTBertDataLoader(DefaultDataLoader):
    def _generate_dataloader(self, dataset, batch_size, last_batch, collate_fn,
                             sampler, batch_sampler, num_workers, pin_memory,
                             shuffle, distributed):
        import numpy as np
        from torch.utils.data import DataLoader, SequentialSampler
        sampler = SequentialSampler(dataset)
        dataloader = DataLoader(dataset, sampler=sampler, \
                                batch_size=batch_size)
        dynamic_length = dataset.dynamic_length
        model_type = dataset.model_type
        max_seq_length = dataset.max_seq_length

        for batch in dataloader:
            try:
                batch_seq_length = max_seq_length if not dynamic_length \
                    else torch.max(batch[-2], 0)[0].item()
                batch = tuple(t.detach().cpu().numpy()  \
                            if not isinstance(t, np.ndarray) else t \
                            for t in batch)
                if model_type == 'bert':
                    data = [
                        batch[0][:,:batch_seq_length],
                        batch[1][:,:batch_seq_length],
                        batch[2][:,:batch_seq_length]
                    ]
                else:
                    data = [
                        batch[0][:,:batch_seq_length],
                        batch[1][:,:batch_seq_length]
                    ]
                label = batch[-1]
                yield data, label
            except StopIteration:
                return

class ONNXRTDataLoader(BaseDataLoader):
    def _generate_dataloader(self, dataset, batch_size, last_batch, collate_fn,
                             sampler, batch_sampler, num_workers, pin_memory,
                             shuffle, distributed):
        if shuffle:
            logging.warning('Shuffle is not supported yet in ONNXRTDataLoader, ' \
                            'ignoring shuffle keyword.')

        if isinstance(dataset, ONNXRTBertDataset):
            return ONNXRTBertDataLoader(dataset, batch_size, last_batch, collate_fn,
                                     sampler, batch_sampler, num_workers, pin_memory,
                                     shuffle, distributed)
        else:
            return DefaultDataLoader(dataset, batch_size, last_batch, collate_fn,
                                     sampler, batch_sampler, num_workers, pin_memory,
                                     shuffle, distributed)



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List
from .. import globals
from ..utils.line_operation import get_line_indent_level
from ..utils.line_operation import multi_line_comment_detection
from ..utils.line_operation import single_line_comment_or_empty_line_detection
import pprint


class CodeLine:
    def __init__(self):
        self.file_path = None
        self.line_idx = None
        self.line_content = None
        self.indent_level = None
        self.is_multi_line_comment = None
        self.is_single_line_comment_or_empty = None
        self.is_class_def_line = None
        self.is_in_class = None
        self.class_name = None
        self.parent_class_name = None
        self.class_def_line_idx = None
        self.class_end_line_idx = None
        self.is_func_def_line = None
        self.is_in_func = None
        self.func_name = None
        self.func_return_idx = None
        self.return_item = None
        self.func_def_line_idx = None
        self.func_end_line_idx = None

    def print_info(self):
        pp = pprint.PrettyPrinter()
        pp.pprint(self.__dict__)


def register_code_line():
    print_class_related_info = False
    print_func_related_info = False
    print("{:<100} {:<10} {:<20} {:<20} {:<20} {:<40} {:<20} {:<20}".format('line',
                                                                            'line_idx',
                                                                            'is_class_def_line',
                                                                            'is_in_class',
                                                                            'class_name',
                                                                            'parent_class_name',
                                                                            'class_def_line_idx',
                                                                            'class_end_line_idx')) \
        if print_class_related_info else 0
    print("{:<100} {:<10} {:<20} {:<20} {:<20} {:<20} {:<20} {:<20} {:<20}".format('line',
                                                                                   'line_idx',
                                                                                   'is_func_def_line',
                                                                                   'is_in_func',
                                                                                   'func_name',
                                                                                   'func_return_idx',
                                                                                   'return_item',
                                                                                   'func_def_line_idx',
                                                                                   'func_end_line_idx')) \
        if print_func_related_info else 0
    for path in globals.list_code_path:
        code = open(path, 'r').read()
        lines = code.split('\n')

        line_idx = 0
        is_multi_line_comment = False
        end_multi_line_comment_flag = False

        is_class_def_line = False
        is_in_class = False
        class_name = ""
        parent_class_name = []
        class_def_line_idx = -1
        class_end_line_idx = -1

        is_func_def_line = False
        is_in_func = False
        func_name = ""
        func_return_idx = -1
        return_item = ""
        func_def_line_idx = -1
        func_end_line_idx = -1

        for line in lines:
            CL = CodeLine()
            CL.file_path = path
            CL.line_idx = line_idx
            CL.line_content = line
            CL.indent_level = get_line_indent_level(line)

            is_multi_line_comment, end_multi_line_comment_flag = multi_line_comment_detection(
                line, is_multi_line_comment, end_multi_line_comment_flag)
            CL.is_multi_line_comment = is_multi_line_comment

            is_single_line_comment_or_empty = single_line_comment_or_empty_line_detection(
                line)
            CL.is_single_line_comment_or_empty = is_single_line_comment_or_empty

            # class
            is_class_def_line = False
            if "class " in line and line.lstrip()[0:5] == "class":
                is_in_class = True
                is_class_def_line = True
                line_ls = line.lstrip()
                if "(" in line_ls:  # "class A(B):"
                    class_name = line_ls[line_ls.find(" ")+1:line_ls.find("(")]
                    parent_content = line_ls[line_ls.find(
                        "(")+1:line_ls.find(")")]
                    if "," in parent_content:  # "class A(B, C):"
                        parent_class_name = []
                        parent_content_items = parent_content.split(", ")
                        for parent_content_item in parent_content_items:
                            parent_class_name.append(parent_content_item)
                    else:  # "class A(B):"
                        parent_class_name = [parent_content]
                else:  # "class A:"
                    class_name = line_ls[line_ls.find(" ")+1:line_ls.find(":")]
                    parent_class_name = []

                # search for class end line
                class_def_indent_level = get_line_indent_level(line)
                class_def_line_idx = line_idx
                search_idx = line_idx + 1
                search_following_lines = True
                _is_multi_line_comment = False
                _end_multi_line_comment_flag = False
                while search_following_lines:
                    try:
                        following_line = lines[search_idx]
                    except:  # end of file situation
                        class_end_line_idx = search_idx
                        break
                    following_indent_level = get_line_indent_level(
                        following_line)

                    _is_multi_line_comment, _end_multi_line_comment_flag = multi_line_comment_detection(
                        following_line, _is_multi_line_comment, _end_multi_line_comment_flag)
                    _is_single_line_comment_or_empty = single_line_comment_or_empty_line_detection(
                        following_line)

                    # judge_1: indent is equal to def indent
                    judge_1 = following_indent_level <= class_def_indent_level
                    # judge_2: not starting with")"
                    judge_2 = True if (
                        following_line != "" and following_line[following_indent_level] != ")") else False
                    # judge_3: is not a comment or empty line
                    judge_3 = True if (
                        not _is_multi_line_comment and not _is_single_line_comment_or_empty) else False

                    if judge_1 and judge_2 and judge_3:
                        search_following_lines = False
                        class_end_line_idx = search_idx

                    search_idx += 1

            if is_in_class and line_idx == class_end_line_idx:
                is_in_class = False
                class_name = ""
                parent_class_name = []
                class_def_line_idx = -1
                class_end_line_idx = -1

            # function
            if is_in_func and line_idx == func_end_line_idx:
                is_in_func = False
                func_return_idx = -1
                return_item = ""
                func_name = ""
                func_def_line_idx = -1
                func_end_line_idx = -1

            is_func_def_line = False
            # only consider outermost function, not consider def(def())
            if not is_in_func and "def " in line:
                is_in_func = True
                is_func_def_line = True
                func_name = line[line.find("def")+4:line.find("(")]

                # search for func end line
                func_def_indent_level = get_line_indent_level(line)
                func_def_line_idx = line_idx
                search_idx = line_idx + 1
                search_following_lines = True
                _is_multi_line_comment = False
                _end_multi_line_comment_flag = False
                while search_following_lines:
                    try:
                        following_line = lines[search_idx]
                    except:  # end of file situation
                        func_end_line_idx = search_idx
                        break
                    following_indent_level = get_line_indent_level(
                        following_line)

                    if "return" in following_line:
                        func_return_idx = search_idx
                        return_item = following_line[following_line.find(
                            "return")+7:].strip()

                    _is_multi_line_comment, _end_multi_line_comment_flag = multi_line_comment_detection(
                        following_line, _is_multi_line_comment, _end_multi_line_comment_flag)
                    _is_single_line_comment_or_empty = single_line_comment_or_empty_line_detection(
                        following_line)

                    # judge_1: indent is equal to def indent
                    judge_1 = following_indent_level <= func_def_indent_level
                    # judge_2: not starting with")"
                    judge_2 = True if (
                        following_line != "" and following_line[following_indent_level] != ")") else False
                    # judge_3: is not a comment or empty line
                    judge_3 = True if (
                        not _is_multi_line_comment and not _is_single_line_comment_or_empty) else False

                    if judge_1 and judge_2 and judge_3:
                        search_following_lines = False
                        func_end_line_idx = search_idx

                    search_idx += 1

            CL.is_class_def_line = is_class_def_line
            CL.is_in_class = is_in_class
            CL.class_name = class_name
            CL.parent_class_name = parent_class_name
            CL.class_def_line_idx = class_def_line_idx
            CL.class_end_line_idx = class_end_line_idx
            print("{:<100} {:<10} {:<20} {:<20} {:<20} {:<40} {:<20} {:<20}".format(line[0:100],
                                                                                    line_idx,
                                                                                    is_class_def_line,
                                                                                    is_in_class,
                                                                                    class_name,
                                                                                    str(
                parent_class_name),
                class_def_line_idx,
                class_end_line_idx)) if print_class_related_info else 0

            CL.is_func_def_line = is_func_def_line
            CL.is_in_func = is_in_func
            CL.func_name = func_name
            CL.func_return_idx = func_return_idx
            CL.return_item = return_item
            CL.func_def_line_idx = func_def_line_idx
            CL.func_end_line_idx = func_end_line_idx
            print("{:<100} {:<10} {:<20} {:<20} \
                {:<20} {:<20} {:<20} {:<20} {:<20}".format(line[0:100],
                                                           line_idx,
                                                           is_func_def_line,
                                                           is_in_func,
                                                           func_name,
                                                           func_return_idx,
                                                           return_item[0:20],
                                                           func_def_line_idx,
                                                           func_end_line_idx)) \
                if print_func_related_info else 0

            globals.list_code_line_instance.append(CL)
            line_idx += 1

    # for i in globals.list_code_line_instance:
    #     i.print_info()



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List
from ..utils.line_operation import get_line_indent_level
from .. import globals
import logging

logging.basicConfig(level=globals.logging_level,
                    format='%(asctime)s %(levelname)s %(message)s',
                    datefmt='%a, %d %b %Y %H:%M:%S +0000')
logger = logging.getLogger(__name__)


# register all relationships of ( [function name] : [return_item] ) pair of the list of code path provided
# but only for "return xxx()" (return a function w/o class prefix) or "return xxx" (return an instance)
# e.g.
# def a1():
#     return b1()
# def b1():
#     return x
# def c():
#     return T.q()
# INPUT: ["example.py"] (above code snippet)
# OUTPUT:
# globals.list_all_function_return_item = ["b1", "x"]
# globals.list_all_function_name = ["a1", "b1"]
def register_func_wrap_pair():
    logger.info(
        f"Analyzing function wrapping relationship for call graph analysis...")
    for path in globals.list_code_path:
        code = open(path, 'r').read()
        lines = code.split('\n')
        line_idx = 0
        is_in_function = False
        func_end_line_idx = -1
        function_def_line_idx = -1
        for line in lines:
            indent_level = get_line_indent_level(line)

            # handle function's end line
            if is_in_function and line_idx == func_end_line_idx:
                is_in_function = False

            # handle function's defnition line, to initiate a function
            if not is_in_function and "def " in line:  # only deal with outermost def
                function_name = line[line.find("def")+4:line.find("(")]

                def_indent_level = get_line_indent_level(line)
                function_def_line_idx = line_idx

                is_in_function = True

                # search for function end line
                search_idx = line_idx + 1
                search_following_lines = True
                multi_comment_flag = False
                while search_following_lines:
                    try:
                        following_line = lines[search_idx]
                    except:  # end of file
                        func_end_line_idx = search_idx
                        break
                    following_indent_level = get_line_indent_level(
                        following_line)

                    # judge_1: indent is equal to def indent
                    judge_1 = following_indent_level <= def_indent_level
                    # judge_2: not starting with")"
                    judge_2 = True if (
                        following_line != "" and following_line[following_indent_level] != ")") else False
                    # judge_3: is not a comment
                    c1 = False
                    c2 = False
                    if multi_comment_flag:
                        c1 = True  # multi-line comment
                    if len(line) > 0 and len(line.lstrip()) > 0 and line.lstrip()[0] == "#":
                        c2 = True  # single-line comment
                    if '"""' in following_line:
                        multi_comment_flag = not multi_comment_flag

                    judge_3 = True if (not c1 and not c2) else False

                    if judge_1 and judge_2 and judge_3:
                        search_following_lines = False
                        func_end_line_idx = search_idx

                    search_idx += 1

                line_idx += 1
                continue

            # handle inside a function
            if is_in_function and line_idx < func_end_line_idx:
                # handle return
                if "return" in line:
                    line_s = line[line.find("return")+7:].strip()
                    # line_s common case: 1. "" 2. "xxx" 3. "xxx, xxx" 3. "xxx()" 4. "xxx(xxx)" 5. "xxx(xxx, xxx)"
                    if line_s == "":  # case 1
                        pass
                    elif line.strip()[0:6] != "return":
                        pass
                    elif 'f"' in line or "#" in line or "if" in line or "." in line or '""' in line or "+" in line:
                        pass
                    elif "(" in line_s:  # case 4 or case 5
                        return_item = line_s[:line_s.find("(")]
                        globals.list_all_function_return_item.append(
                            return_item)
                        globals.list_all_function_name.append(function_name)
                    elif ", " in line_s:  # case 3
                        ls = line_s.split(", ")
                        for return_item in ls:
                            globals.list_all_function_return_item.append(
                                return_item)
                            globals.list_all_function_name.append(
                                function_name)
                    else:  # case 2
                        return_item = line_s
                        globals.list_all_function_return_item.append(
                            return_item)
                        globals.list_all_function_name.append(function_name)

            line_idx += 1
            continue

    logger.debug(
        f"globals.list_all_function_name: {globals.list_all_function_name}")
    logger.debug(
        f"globals.list_all_function_return_item: {globals.list_all_function_return_item}")

# get all wrapper children names of the base function name
# e.g.
# class Net(nn.Module):
#    xxx
#
# def _resnet():
#     model = Net()
#     return model
#
# def resnet34():
#     xxx
#     return _resnet()
#
# def resnet18():
#     xxx
#     return _resnet()
#
# def resnet18_large():
#     xxx
#     return resnet18()
#
# INPUT: "_resnet"
# OUTPUT: ["resnet18", "resnet34", "resnet18_large"]


def get_all_wrap_children(base_function_name: str) -> List:
    length = range(len(globals.list_all_function_return_item))
    base_function_name = [base_function_name]
    do_search = True
    list_child_all = []

    while do_search:
        current_count = len(list_child_all)
        for this_base in base_function_name:
            this_list_child = []
            for i in length:
                if globals.list_all_function_return_item[i] == this_base:
                    this_list_child.append(globals.list_all_function_name[i])
            base_function_name = this_list_child

            list_child_all += this_list_child
            list_child_all = list(set(list_child_all))

        if len(list_child_all) == current_count:
            do_search = False

    return list_child_all



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import onnx
from .base_operator import QuantOperatorBase
from onnxruntime.quantization.quant_utils import QuantizedValueType, \
        attribute_to_kwarg, ms_domain
from onnx import onnx_pb as onnx_proto
from neural_compressor.adaptor.ox_utils.util import QuantizedValue
from .qdq_base_operator import QDQOperatorBase

class QDQConcat(QDQOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def quantize(self):
        node = self.node
        inits = [i.name for i in self.quantizer.model.initializer()]
        if all([inp not in self.quantizer.quantized_value_map and inp not in inits \
            for inp in node.input]) or \
            not all([inp in self.quantizer.quantized_value_map or inp in inits \
            for inp in node.input]):
            return
        for idx, inp in enumerate(node.input):
            initializer_use_weight_qType = inp not in inits
            self.quantizer.quantize_inputs(node, [idx], initializer_use_weight_qType)
        if not self.disable_qdq_for_node_output or self.quantizer.mode != 'qdq':
            self.quantizer.quantize_outputs(node)
        node.name = node.name + "_quant"

class QLinearConcat(QuantOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def convert(self):
        node = self.node

        parents = self.quantizer.model.get_parents(node)
        children = self.quantizer.model.get_children(node)
        if len(children) == 0 or len(parents) == 0:
            return

        if all([i.op_type == 'DequantizeLinear' for i in parents]) and \
            any([i.op_type == 'QuantizeLinear' for i in children]):
            inputs = []

            inputs.extend([i for i in children if i.op_type == 'QuantizeLinear'][0].input[1:])
            for parent in parents:
                inputs.extend(parent.input)
                self.quantizer.remove_nodes.append(parent)
            for child in children:
                if child.op_type == 'QuantizeLinear':
                    self.quantizer.remove_nodes.append(child)
                    self.quantizer.model.replace_input_of_all_nodes(
                        child.output[0], node.output[0] + '_quantized')
            
            kwargs = {}
            for attribute in node.attribute:
                kwargs.update(attribute_to_kwarg(attribute))
            kwargs["domain"] = ms_domain
            qlconcat_node = onnx.helper.make_node("QLinearConcat", 
                                                  inputs, 
                                                  [node.output[0] + "_quantized"], 
                                                  node.name, 
                                                  **kwargs)        

            self.quantizer.new_nodes += [qlconcat_node]
            self.quantizer.remove_nodes.append(node)



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ...utils.line_operation import get_line_lhs


class TensorFlowKerasAMP(object):
    def __init__(self, file) -> None:
        self.file = file
        self.result = []
        self.keras_edited_flag = False

    def transform(self):
        # import pdb
        # pdb.set_trace()
        lines = self.file.split('\n')
        for line in lines:
            if self.is_modify(line):
                if '.ConfigProto()' in line:  # TF AMP
                    config_name = get_line_lhs(line)
                    new_line_1 = "from tensorflow.core.protobuf import rewriter_config_pb2"
                    new_line_2 = config_name + \
                        ".graph_options.rewrite_options.auto_mixed_precision_mkl = \
                            rewriter_config_pb2.RewriterConfig.ON"
                    self.result.append(line)
                    self.result.append(new_line_1)
                    self.result.append(new_line_2)
                elif 'keras' in line and 'import' in line:  # Keras AMP
                    if not self.keras_edited_flag:
                        new_line_1 = "from tensorflow.keras.mixed_precision import experimental as mixed_precision"
                        new_line_2 = "policy = mixed_precision.Policy('mixed_bfloat16')"
                        new_line_3 = "mixed_precision.set_policy(policy)"
                        self.result.append(line)
                        self.result.append(new_line_1)
                        self.result.append(new_line_2)
                        self.result.append(new_line_3)
                        self.keras_edited_flag = True
                    else:
                        self.result.append(line)
            else:

                self.result.append(line)
        for index, line in enumerate(self.result):
            if index != len(self.result)-1:
                self.result[index] += '\n'
        return ''.join(self.result)

    def is_modify(self, s):
        if '.ConfigProto()' in s or ('keras' in s and 'import' in s):
            return True
        else:
            return False



#
#  -*- coding: utf-8 -*-
#
#  Copyright (c) 2021 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.core.framework import attr_value_pb2
from tensorflow.core.framework import node_def_pb2
from tensorflow.python.framework import dtypes
from .graph_transform_base import GraphTransformBase


class InsertLogging(GraphTransformBase):
    """
    Insert logging graph transformation.
    """
    op_output_type_mapping = {
        "RequantizationRange":
        [dtypes.float32.as_datatype_enum, dtypes.float32.as_datatype_enum],
        "RequantizationRangePerChannel":
        [dtypes.float32.as_datatype_enum, dtypes.float32.as_datatype_enum],
        "QuantizedConv2DWithBiasAndRelu": [dtypes.qint32.as_datatype_enum],
        "QuantizedConv2DWithBiasAndReluAndRequantize":
        [dtypes.quint8.as_datatype_enum],
        "QuantizedConv2DWithBiasAndRequantize":
        [dtypes.qint8.as_datatype_enum],
        "QuantizedConv2DWithBiasSignedSumAndReluAndRequantize":
        [dtypes.qint8.as_datatype_enum],
        "QuantizedConv2DWithBiasSumAndReluAndRequantize":
        [dtypes.quint8.as_datatype_enum],
        "QuantizedDepthwiseConv2DWithBiasAndReluAndRequantize":
        [dtypes.quint8.as_datatype_enum],
        "QuantizedConv2DWithBias": [dtypes.qint32.as_datatype_enum],
        "Relu": [dtypes.float32.as_datatype_enum],
        "Relu6": [dtypes.float32.as_datatype_enum],
        "AvgPool": [dtypes.float32.as_datatype_enum],
        "MaxPool": [dtypes.float32.as_datatype_enum],
        "BiasAdd": [dtypes.float32.as_datatype_enum],
        "Max": [dtypes.float32.as_datatype_enum],
        "Min": [dtypes.float32.as_datatype_enum]
    }

    def __init__(self,
                 input_pb,
                 ops=[],
                 node_name_list=[],
                 show_name=True,
                 show_op=False,
                 first_n=-1,
                 summarize=1024,
                 message="",
                 dump_fp32=False):
        super(InsertLogging, self).__init__(input_pb)

        self.parse_input_pb()
        self.ops = ops
        self.node_name_list = node_name_list
        self.show_name = show_name
        self.show_op = show_op
        self.message = message
        self.first_n = first_n
        self.summarize = summarize
        self.output_name_index_mapping = {}
        self.input_rename = {}
        self.dump_fp32 = dump_fp32

    def _get_suffix(self, input_str):
        """
        Split the node name into two parts.
        Returns:
            Pure string name without suffix
            Index of the node
        """
        splitted_str = input_str.split(':')
        if len(splitted_str) < 2:
            return input_str, 0

        return splitted_str[0], int(splitted_str[-1])

    def _get_output_index_mapping(self):
        """
        Get the output_node_name and index mapping.
        """
        for node_name in self.node_mapping:
            for node_input in self.node_mapping[node_name].input:
                node_stripped_name, suffix = self._get_suffix(node_input)

                if node_stripped_name not in self.output_name_index_mapping:
                    self.output_name_index_mapping[node_stripped_name] = set()

                self.output_name_index_mapping[node_stripped_name].add(suffix)

    def _insert_node(self):
        """
        Insert the Print OP into the graph
        """
        for node_name in self.node_mapping:
            if node_name not in self.output_name_index_mapping or (
                    not self.dump_fp32 and node_name.find("eightbit") == -1):
                continue

            if self.ops and self.node_mapping[
                    node_name].op in self.ops or node_name in self.node_name_list:
                name_suffix = "__print__"
                print_node = node_def_pb2.NodeDef()
                print_node.op = "Print"
                print_node.name = node_name + name_suffix
                node_message = ''
                if self.show_op:
                    node_message += ';' + self.node_mapping[node_name].op + ';'
                if self.show_name:
                    node_message += ';' + print_node.name + ';'
                node_message += self.message

                print_node.attr["message"].s = node_message.encode()
                print_node.attr["first_n"].i = self.first_n
                print_node.attr["summarize"].i = self.summarize

                print_node.input.append(node_name + ":0")
                print_node.attr["T"].CopyFrom(
                    attr_value_pb2.AttrValue(type=self.op_output_type_mapping[
                        self.node_mapping[node_name].op][0]))

                if self.node_mapping[node_name].op in (
                        "QuantizedConv2DWithBias",
                        "QuantizedConv2DWithBiasAndRelu"):

                    for index in sorted(
                            self.output_name_index_mapping[node_name])[:1]:
                        print_node.input.append(node_name + ":" + str(index))

                        print_node_1 = node_def_pb2.NodeDef()
                        print_node_1.op = "Print"
                        print_node_1.name = node_name + name_suffix + "_min_output"

                        print_node_1.attr["message"].s = (
                            node_message + "_min_output").encode()
                        print_node_1.attr["first_n"].i = self.first_n
                        print_node_1.attr["summarize"].i = self.summarize
                        print_node_1.attr["U"].list.CopyFrom(
                            attr_value_pb2.AttrValue.ListValue(
                                type=[dtypes.float32.as_datatype_enum]))
                        print_node_1.attr["T"].CopyFrom(
                            attr_value_pb2.AttrValue(
                                type=dtypes.float32.as_datatype_enum))
                        print_node_1.input.append(node_name + ":1")
                        print_node_1.input.append(node_name + ":1")
                        self.input_graph.node.extend([print_node_1])
                        self.input_rename[node_name +
                                          ':1'] = print_node_1.name + ':0'

                        print_node_2 = node_def_pb2.NodeDef()
                        print_node_2.op = "Print"
                        print_node_2.name = node_name + name_suffix + "_max_output"

                        print_node_2.attr["message"].s = (
                            node_message + "_max_output").encode()
                        print_node_2.attr["first_n"].i = self.first_n
                        print_node_2.attr["summarize"].i = self.summarize
                        print_node_2.attr["U"].list.CopyFrom(
                            attr_value_pb2.AttrValue.ListValue(
                                type=[dtypes.float32.as_datatype_enum]))
                        print_node_2.attr["T"].CopyFrom(
                            attr_value_pb2.AttrValue(
                                type=dtypes.float32.as_datatype_enum))
                        print_node_2.input.append(node_name + ":2")
                        print_node_2.input.append(node_name + ":2")
                        self.input_graph.node.extend([print_node_2])
                        self.input_rename[node_name +
                                          ':2'] = print_node_2.name + ':0'
                else:
                    for index in range(
                            len(self.op_output_type_mapping[
                                self.node_mapping[node_name].op])):
                        print_node.input.append(node_name + ":" + str(
                            sorted(self.output_name_index_mapping[node_name])
                            [index]))

                print_node.attr["U"].list.CopyFrom(
                    attr_value_pb2.AttrValue.ListValue(
                        type=self.op_output_type_mapping[
                            self.node_mapping[node_name].op]))

                self.input_graph.node.extend([print_node])

                self.input_rename[node_name + ':0'] = print_node.name + ':0'

    def _rename_node(self):
        """
        Rename the original input node and connect to new added print node.
        """
        for node_name in self.node_mapping:
            for index, input_name in enumerate(
                    self.node_mapping[node_name].input):
                if input_name in self.input_rename:
                    self.node_mapping[node_name].input[
                        index] = self.input_rename[input_name]
                elif input_name + ':0' in self.input_rename:
                    self.node_mapping[node_name].input[
                        index] = self.input_rename[input_name + ':0']

    def do_transformation(self):
        """
        Execute the insert logging transformation.
        :return: Transformed graph
        """
        self._get_output_index_mapping()
        self._insert_node()
        self._rename_node()

        return self.input_graph



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from collections import OrderedDict
import numpy as np
from .strategy import strategy_registry, TuneStrategy


@strategy_registry
class MSETuneStrategy(TuneStrategy):
    """The tuning strategy using MSE policy in tuning space.

       This MSE policy runs fp32 model and int8 model seperately to get all activation tensors,
       and then compares those tensors by MSE algorithm to order all ops with MSE distance for
       deciding the impact of each op to final accuracy.
       It will be used to define opwise tuningspace by priority.

    Args:
        model (object):                        The FP32 model specified for low precision tuning.
        conf (Conf):                           The Conf class instance initialized from user yaml
                                               config file.
        q_dataloader (generator):              Data loader for calibration, mandatory for
                                               post-training quantization.
                                               It is iterable and should yield a tuple (input,
                                               label) for calibration dataset containing label,
                                               or yield (input, _) for label-free calibration
                                               dataset. The input could be a object, list, tuple or
                                               dict, depending on user implementation, as well as
                                               it can be taken as model input.
        q_func (function, optional):           Reserved for future use.
        eval_dataloader (generator, optional): Data loader for evaluation. It is iterable
                                               and should yield a tuple of (input, label).
                                               The input could be a object, list, tuple or dict,
                                               depending on user implementation, as well as it can
                                               be taken as model input. The label should be able
                                               to take as input of supported metrics. If this
                                               parameter is not None, user needs to specify
                                               pre-defined evaluation metrics through configuration
                                               file and should set "eval_func" parameter as None.
                                               Tuner will combine model, eval_dataloader and
                                               pre-defined metrics to run evaluation process.
        eval_func (function, optional):        The evaluation function provided by user.
                                               This function takes model as parameter, and
                                               evaluation dataset and metrics should be
                                               encapsulated in this function implementation and
                                               outputs a higher-is-better accuracy scalar value.

                                               The pseudo code should be something like:

                                               def eval_func(model):
                                                    input, label = dataloader()
                                                    output = model(input)
                                                    accuracy = metric(output, label)
                                                    return accuracy
        dicts (dict, optional):                The dict containing resume information.
                                               Defaults to None.

    """

    def __init__(self, model, conf, dataloader, q_func=None,
                 eval_dataloader=None, eval_func=None, dicts=None, q_hooks=None):
        self.ordered_ops = None
        super().__init__(
            model,
            conf,
            dataloader,
            q_func,
            eval_dataloader,
            eval_func,
            dicts,
            q_hooks)

    def __getstate__(self):
        for history in self.tuning_history:
            if self._same_yaml(history['cfg'], self.cfg):
                history['ordered_ops'] = self.ordered_ops
        save_dict = super().__getstate__()
        return save_dict

    def mse_metric_gap(self, fp32_tensor, dequantize_tensor):
        """Calculate the euclidean distance between fp32 tensor and int8 dequantize tensor

        Args:
            fp32_tensor (tensor): The FP32 tensor.
            dequantize_tensor (tensor): The INT8 dequantize tensor.
        """
        fp32_max = np.max(fp32_tensor)
        fp32_min = np.min(fp32_tensor)
        dequantize_max = np.max(dequantize_tensor)
        dequantize_min = np.min(dequantize_tensor)
        fp32_tensor = (fp32_tensor - fp32_min) / (fp32_max - fp32_min)
        dequantize_tensor = (dequantize_tensor - dequantize_min) / \
            (dequantize_max - dequantize_min)
        diff_tensor = fp32_tensor - dequantize_tensor
        euclidean_dist = np.sum(diff_tensor ** 2)
        return euclidean_dist / fp32_tensor.size

    def next_tune_cfg(self):
        """The generator of yielding next tuning config to traverse by concrete strategies
           according to last tuning result.

        """
        # Model wise tuning
        op_cfgs = {}
        best_cfg = None
        if len(self.metric_name) == 1 or self.metric_weight is not None:
            best_acc = float('-inf') if self.higher_is_better else float('inf')
        else:
            best_acc = [float('-inf') if higher_is_better else float('inf') for \
                higher_is_better in self.metric_criterion]

        for i, iterations in enumerate(self.calib_iter):
            op_cfgs['calib_iteration'] = int(iterations)
            op_cfgs['calib_sampling_size'] = int(self.calib_sampling_size[i])
            for combined_cfg in self.combined_model_wise_quant_cfgs:
                op_cfgs['op'] = OrderedDict()
                for op, op_cfg in self.opwise_quant_cfgs.items():
                    if op[1] in combined_cfg.keys() and len(op_cfg) > 0:
                        op_cfgs['op'][op] = copy.deepcopy(
                            self._get_common_cfg(combined_cfg[op[1]], op_cfg))
                    else:
                        op_cfgs['op'][op] = copy.deepcopy(
                            self.opwise_tune_cfgs[op][0])

                yield op_cfgs
                acc, _ = self.last_tune_result

                if not isinstance(acc, list) and ((self.higher_is_better and acc >= best_acc) \
                    or (not self.higher_is_better and acc <= best_acc)):
                    best_acc = acc
                    best_cfg = copy.deepcopy(op_cfgs)
                elif len(self.metric_name) > 1 and self.metric_weight is not None:
                    acc = np.mean(np.array(acc) * self.metric_weight)
                    if (self.higher_is_better and acc >= best_acc) or \
                        (not self.higher_is_better and acc <= best_acc):
                        best_acc = acc
                        best_cfg = copy.deepcopy(op_cfgs)
                elif len(self.metric_name) > 1 and self.metric_weight is None:
                    if all([acc_i >= best_i if higher_is_better else acc_i <= best_i for \
                        acc_i, best_i, higher_is_better in \
                        zip(acc, best_acc, self.metric_criterion)]):
                        best_acc = acc
                        best_cfg = copy.deepcopy(op_cfgs)

        if best_cfg is not None:
            # Inspect FP32 and dequantized tensor
            if self.ordered_ops is None or "ops_mse" not in locals():
                op_lists = self.opwise_quant_cfgs.keys()
                op_mapping = {}
                for (op_name, op_type) in list(op_lists):
                    op_mapping[op_name] = (op_name, op_type)
                op_lists = [op_name for (op_name, op_type) in list(op_lists)]
                fp32_dump_content = self.adaptor.inspect_tensor(self.model, 
                    self.calib_dataloader, op_lists, [1], inspect_type='activation', 
                    save_to_disk=True, save_path="./nc_workspace/", 
                    quantization_cfg=best_cfg)
                fp32_tensor_dict = fp32_dump_content['activation'][0]
                best_qmodel = self.adaptor.quantize(best_cfg, self.model, self.calib_dataloader)
                quant_dump_content = self.adaptor.inspect_tensor(best_qmodel, 
                    self.calib_dataloader, op_lists, [1], inspect_type='activation',
                    save_to_disk=True, save_path="./nc_workspace/", 
                    quantization_cfg=best_cfg)
                dequantize_tensor_dict = quant_dump_content['activation'][0]
                ops_mse = {
                    op: self.mse_metric_gap(
                        list(fp32_tensor_dict[op].values())[0],
                        list(dequantize_tensor_dict[op].values())[0]) for op in fp32_tensor_dict}
                self.ordered_ops = sorted(ops_mse.keys(), key=lambda key: ops_mse[key],
                                          reverse=self.higher_is_better)

            if ops_mse is not None:
                ordered_ops = sorted(ops_mse.keys(), key=lambda key: ops_mse[key], \
                                     reverse=self.higher_is_better)
                op_cfgs = copy.deepcopy(best_cfg)
                for op in ordered_ops:
                    if not isinstance(op, tuple):
                        cfg_key = [item[0] for item in list(op_cfgs['op'].keys())]
                        op = list(op_cfgs['op'].keys())[cfg_key.index(op)]
                    old_cfg = copy.deepcopy(op_cfgs['op'][op])
                    op_cfgs['op'][op]['activation'].clear()
                    op_cfgs['op'][op]['activation']['dtype'] = 'fp32'
                    if 'weight' in op_cfgs['op'][op] and op_cfgs['op'][op]['weight'] is not None:
                        op_cfgs['op'][op]['weight'].clear()
                        op_cfgs['op'][op]['weight']['dtype'] = 'fp32'
                    yield op_cfgs
                    acc, _ = self.last_tune_result

                    if not isinstance(acc, list):
                        if ((self.higher_is_better and acc <= best_acc) \
                            or (not self.higher_is_better and acc >= best_acc)):
                            op_cfgs['op'][op] = copy.deepcopy(old_cfg)
                        else:
                            best_acc = acc
                    elif len(self.metric_name) > 1 and self.metric_weight is not None:
                        acc = np.mean(np.array(acc) * self.metric_weight)
                        if (self.higher_is_better and acc <= best_acc) or \
                            (not self.higher_is_better and acc >= best_acc):
                            op_cfgs['op'][op] = copy.deepcopy(old_cfg)
                        else:
                            best_acc = acc
                    elif len(self.metric_name) > 1 and self.metric_weight is None:
                        if all([acc_i >= best_i if higher_is_better else acc_i <= best_i for \
                            acc_i, best_i, higher_is_better in \
                            zip(acc, best_acc, self.metric_criterion)]):
                            op_cfgs['op'][op] = copy.deepcopy(old_cfg)
                        else:
                            best_acc = acc

                op_cfgs = copy.deepcopy(best_cfg)
                for op in ordered_ops:
                    op = op_mapping[op]
                    op_cfgs['op'][op]['activation'].clear()
                    op_cfgs['op'][op]['activation']['dtype'] = 'fp32'
                    if 'weight' in op_cfgs['op'][op] and op_cfgs['op'][op]['weight'] is not None:
                        op_cfgs['op'][op]['weight'].clear()
                        op_cfgs['op'][op]['weight']['dtype'] = 'fp32'
                    yield op_cfgs
        else:
            op_cfgs['op'] = OrderedDict()
            for op in self.opwise_tune_cfgs.keys():
                op_cfgs['op'][op] = copy.deepcopy(self.opwise_tune_cfgs[op][0])
            yield op_cfgs

        return
